#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Reader/Writer for plain binfiles.
"""

__version__ = "0.1.1"

__copyright__ = """
    objutils - Object file library for Python.

   (C) 2010-2019 by Christoph Schueler <cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

from contextlib import closing
import io
import zipfile

from objutils.section import Section
from objutils.image import Image
from objutils.utils import PYTHON_VERSION, create_string_buffer

##
## TODO: binzipped format: a separate file for each section + MANIFEST (csv: fname, address, length)
##

class NoContiniousError(Exception): pass


class Reader(object):

    def load(self, fp, address = 0x0000):
        if isinstance(fp, str):
            fp = open(fp, "rb")
        data = fp.read()
        sec = Section(address, data)
        img = Image([sec], valid = True)
        if hasattr(fp, "close"):
            fp.close()
        return img

    def loads(self, image, address = 0x0000):
        if PYTHON_VERSION.major == 3:
            if isinstance(image, str):
                return self.load(create_string_buffer(bytes(image, "ascii")), address)
            else:
                return self.load(create_string_buffer(image), address)
        else:
            return self.load(create_string_buffer(image), address)


class Writer(object):
    def dump(self, fp, image, filler = b'\xff', **kws):
        if isinstance(fp, str):
            fp = open(fp, "wb")
        fp.write(self.dumps(image, filler))
        if hasattr(fp, "close"):
            fp.close()

    def dumps(self, image, filler = b'\xff', **kws):
        if not isinstance(filler, (bytes, int)):
            raise TypeError("filler must be of type 'bytes' or 'int'")
        if isinstance(filler, bytes) and len(filler) > 1:
            raise TypeError("filler must be a single byte")
        elif isinstance(filler, int) and filler > 255:
            raise ValueError("filler must be in range 0..255")
        result = bytearray()
        previous_address = None
        previous_length = None

        if hasattr(image, "sections") and  not image.sections:
            return b''
        sections = sorted(image.sections, key = lambda x: x.start_address)
        for section in sections:
            if not previous_address is None:
                gap = section.start_address - (previous_address + previous_length)
                if gap > 0:
                    result.extend(filler * gap)
            result.extend(section.data)
            previous_address = section.start_address
            previous_length = section.length
        return result

#

class BinZipReader(object):
    pass


class BinZipWriter(object):

    SECTION_FILE_NAME = "image{0:d}.bin"
    MANIFEST_FILE_NAME = "IMAGES.mf"

    def dump(self, fp, image, **kws):
        fp.write(self.dumps(image))
        if hasattr(fp, "close"):
            fp.close()

    def dumps(self, image, **kws):

        if hasattr(image, "sections") and  not image.sections:
            return b''
        sections = sorted(image.sections, key = lambda x: x.start_address)
        manifest_buffer = io.StringIO()
        out_buffer = io.BytesIO()
        #out_buffer = io.StringIO()
        print("BUF", out_buffer)
        with closing(zipfile.ZipFile(out_buffer, mode = "w")) as outFile:
            print(outFile)
            for idx, section in enumerate(sections):
                print(section.start_address, section.length)
                #print("FN", BinZipWriter.SECTION_FILE_NAME.format(idx))
                manifest_buffer.write(BinZipWriter.SECTION_FILE_NAME.format(idx))
                manifest_buffer.write("\t")
                manifest_buffer.write(str(section.start_address))
                manifest_buffer.write("\t")
                manifest_buffer.write(str(section.length))
                manifest_buffer.write("\n")
        manifest_buffer.seek(0)
        print(manifest_buffer.read())
        return ''



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__copyright__ = """
    objutils - Object file library for Python.

   (C) 2010-2020 by Christoph Schueler <github.com/Christoph2,
                                        cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import enum

class Base(object):

    def __init__(self, value):
        self._value = value

    def _getValue(self):
        return self._value

    def __str__(self):
        return "< {0!s} >".format((self.MAP.get(self._value, self._value)))

    __repr__ = __str__

    value = property(_getValue)

##
## Tag encodings
##
DW_TAG_array_type               = 0x01
DW_TAG_class_type               = 0x02
DW_TAG_entry_point              = 0x03
DW_TAG_enumeration_type         = 0x04
DW_TAG_formal_parameter         = 0x05
DW_TAG_imported_declaration     = 0x08
DW_TAG_label                    = 0x0a
DW_TAG_lexical_block            = 0x0b
DW_TAG_member                   = 0x0d
DW_TAG_pointer_type             = 0x0f
DW_TAG_reference_type           = 0x10
DW_TAG_compile_unit             = 0x11
DW_TAG_string_type              = 0x12
DW_TAG_structure_type           = 0x13
DW_TAG_subroutine_type          = 0x15
DW_TAG_typedef                  = 0x16
DW_TAG_union_type               = 0x17
DW_TAG_unspecified_parameters   = 0x18
DW_TAG_variant                  = 0x19
DW_TAG_common_block             = 0x1a
DW_TAG_common_inclusion         = 0x1b
DW_TAG_inheritance              = 0x1c
DW_TAG_inlined_subroutine       = 0x1d
DW_TAG_module                   = 0x1e
DW_TAG_ptr_to_member_type       = 0x1f
DW_TAG_set_type                 = 0x20
DW_TAG_subrange_type            = 0x21
DW_TAG_with_stmt                = 0x22
DW_TAG_access_declaration       = 0x23
DW_TAG_base_type                = 0x24
DW_TAG_catch_block              = 0x25
DW_TAG_const_type               = 0x26
DW_TAG_constant                 = 0x27
DW_TAG_enumerator               = 0x28
DW_TAG_file_type                = 0x29
DW_TAG_friend                   = 0x2a
DW_TAG_namelist                 = 0x2b
DW_TAG_namelist_item            = 0x2c
DW_TAG_packed_type              = 0x2d
DW_TAG_subprogram               = 0x2e
DW_TAG_template_type_parameter  = 0x2f
DW_TAG_template_value_parameter = 0x30
DW_TAG_thrown_type              = 0x31
DW_TAG_try_block                = 0x32
DW_TAG_variant_part             = 0x33
DW_TAG_variable                 = 0x34
DW_TAG_volatile_type            = 0x35
DW_TAG_dwarf_procedure          = 0x36
DW_TAG_restrict_type            = 0x37
DW_TAG_interface_type           = 0x38
DW_TAG_namespace                = 0x39
DW_TAG_imported_module          = 0x3a
DW_TAG_unspecified_type         = 0x3b
DW_TAG_partial_unit             = 0x3c
DW_TAG_imported_unit            = 0x3d
DW_TAG_condition                = 0x3f
DW_TAG_shared_type              = 0x40
DW_TAG_type_unit                = 0x41
DW_TAG_rvalue_reference_type    = 0x42
DW_TAG_template_alias           = 0x43
DW_TAG_lo_user                  = 0x4080
DW_TAG_hi_user                  = 0xffff

TAG_MAP = {
    DW_TAG_array_type               : "DW_TAG_array_type",
    DW_TAG_class_type               : "DW_TAG_class_type",
    DW_TAG_entry_point              : "DW_TAG_entry_point",
    DW_TAG_enumeration_type         : "DW_TAG_enumeration_type",
    DW_TAG_formal_parameter         : "DW_TAG_formal_parameter",
    DW_TAG_imported_declaration     : "DW_TAG_imported_declaration",
    DW_TAG_label                    : "DW_TAG_label",
    DW_TAG_lexical_block            : "DW_TAG_lexical_block",
    DW_TAG_member                   : "DW_TAG_member",
    DW_TAG_pointer_type             : "DW_TAG_pointer_type",
    DW_TAG_reference_type           : "DW_TAG_reference_type",
    DW_TAG_compile_unit             : "DW_TAG_compile_unit",
    DW_TAG_string_type              : "DW_TAG_string_type",
    DW_TAG_structure_type           : "DW_TAG_structure_type",
    DW_TAG_subroutine_type          : "DW_TAG_subroutine_type",
    DW_TAG_typedef                  : "DW_TAG_typedef",
    DW_TAG_union_type               : "DW_TAG_union_type",
    DW_TAG_unspecified_parameters   : "DW_TAG_unspecified_parameters",
    DW_TAG_variant                  : "DW_TAG_variant",
    DW_TAG_common_block             : "DW_TAG_common_block",
    DW_TAG_common_inclusion         : "DW_TAG_common_inclusion",
    DW_TAG_inheritance              : "DW_TAG_inheritance",
    DW_TAG_inlined_subroutine       : "DW_TAG_inlined_subroutine",
    DW_TAG_module                   : "DW_TAG_module",
    DW_TAG_ptr_to_member_type       : "DW_TAG_ptr_to_member_type",
    DW_TAG_set_type                 : "DW_TAG_set_type",
    DW_TAG_subrange_type            : "DW_TAG_subrange_type",
    DW_TAG_with_stmt                : "DW_TAG_with_stmt",
    DW_TAG_access_declaration       : "DW_TAG_access_declaration",
    DW_TAG_base_type                : "DW_TAG_base_type",
    DW_TAG_catch_block              : "DW_TAG_catch_block",
    DW_TAG_const_type               : "DW_TAG_const_type",
    DW_TAG_constant                 : "DW_TAG_constant",
    DW_TAG_enumerator               : "DW_TAG_enumerator",
    DW_TAG_file_type                : "DW_TAG_file_type",
    DW_TAG_friend                   : "DW_TAG_friend",
    DW_TAG_namelist                 : "DW_TAG_namelist",
    DW_TAG_namelist_item            : "DW_TAG_namelist_item",
    DW_TAG_packed_type              : "DW_TAG_packed_type",
    DW_TAG_subprogram               : "DW_TAG_subprogram",
    DW_TAG_template_type_parameter  : "DW_TAG_template_type_parameter",
    DW_TAG_template_value_parameter : "DW_TAG_template_value_parameter",
    DW_TAG_thrown_type              : "DW_TAG_thrown_type",
    DW_TAG_try_block                : "DW_TAG_thrown_type",
    DW_TAG_variant_part             : "DW_TAG_variant_part",
    DW_TAG_variable                 : "DW_TAG_variable",
    DW_TAG_volatile_type            : "DW_TAG_volatile_type",
    DW_TAG_dwarf_procedure          : "DW_TAG_dwarf_procedure",
    DW_TAG_restrict_type            : "DW_TAG_restrict_type",
    DW_TAG_interface_type           : "DW_TAG_interface_type",
    DW_TAG_namespace                : "DW_TAG_namespace",
    DW_TAG_imported_module          : "DW_TAG_imported_module",
    DW_TAG_unspecified_type         : "DW_TAG_unspecified_type",
    DW_TAG_partial_unit             : "DW_TAG_partial_unit",
    DW_TAG_imported_unit            : "DW_TAG_imported_unit",
    DW_TAG_condition                : "DW_TAG_condition",
    DW_TAG_shared_type              : "DW_TAG_shared_type",
    DW_TAG_type_unit                : "DW_TAG_type_unit",
    DW_TAG_rvalue_reference_type    : "DW_TAG_rvalue_reference_type",
    DW_TAG_template_alias           : "DW_TAG_template_alias",
    DW_TAG_lo_user                  : "DW_TAG_lo_user",
    DW_TAG_hi_user                  : "DW_TAG_hi_user",
}

class Tag(Base):
    MAP = TAG_MAP

##
## Child determination encodings.
##
DW_CHILDREN_no                  = 0x00
DW_CHILDREN_yes                 = 0x01

CHILDREN_MAP = {
    DW_CHILDREN_no              : "DW_CHILDREN_no",
    DW_CHILDREN_yes             : "DW_CHILDREN_yes",
}

class Children(Base):
    MAP = CHILDREN_MAP

##
## Attribute encodings.
##
DW_AT_sibling                   = 0x01
DW_AT_location                  = 0x02
DW_AT_name                      = 0x03
DW_AT_ordering                  = 0x09
DW_AT_byte_size                 = 0x0b
DW_AT_bit_offset                = 0x0c
DW_AT_bit_size                  = 0x0d
DW_AT_stmt_list                 = 0x10
DW_AT_low_pc                    = 0x11
DW_AT_high_pc                   = 0x12
DW_AT_language                  = 0x13
DW_AT_discr                     = 0x15
DW_AT_discr_value               = 0x16
DW_AT_visibility                = 0x17
DW_AT_import                    = 0x18
DW_AT_string_length             = 0x19
DW_AT_common_reference          = 0x1a
DW_AT_comp_dir                  = 0x1b
DW_AT_const_value               = 0x1c
DW_AT_containing_type           = 0x1d
DW_AT_default_value             = 0x1e
DW_AT_inline                    = 0x20
DW_AT_is_optional               = 0x21
DW_AT_lower_bound               = 0x22
DW_AT_producer                  = 0x25
DW_AT_prototyped                = 0x27
DW_AT_return_addr               = 0x2a
DW_AT_start_scope               = 0x2c
DW_AT_bit_stride                = 0x2e
DW_AT_upper_bound               = 0x2f
DW_AT_abstract_origin           = 0x31
DW_AT_accessibility             = 0x32
DW_AT_address_class             = 0x33
DW_AT_artificial                = 0x34
DW_AT_base_types                = 0x35
DW_AT_calling_convention        = 0x36
DW_AT_count                     = 0x37
DW_AT_data_member_location      = 0x38
DW_AT_decl_column               = 0x39
DW_AT_decl_file                 = 0x3a
DW_AT_decl_line                 = 0x3b
DW_AT_declaration               = 0x3c
DW_AT_discr_list                = 0x3d
DW_AT_encoding                  = 0x3e
DW_AT_external                  = 0x3f
DW_AT_frame_base                = 0x40
DW_AT_friend                    = 0x41
DW_AT_identifier_case           = 0x42
DW_AT_macro_info                = 0x43
DW_AT_namelist_item             = 0x44
DW_AT_priority                  = 0x45
DW_AT_segment                   = 0x46
DW_AT_specification             = 0x47
DW_AT_static_link               = 0x48
DW_AT_type                      = 0x49
DW_AT_use_location              = 0x4a
DW_AT_variable_parameter        = 0x4b
DW_AT_virtuality                = 0x4c
DW_AT_vtable_elem_location      = 0x4d
DW_AT_allocated                 = 0x4e
DW_AT_associated                = 0x4f
DW_AT_data_location             = 0x50
DW_AT_byte_stride               = 0x51
DW_AT_entry_pc                  = 0x52
DW_AT_use_UTF8                  = 0x53
DW_AT_extension                 = 0x54
DW_AT_ranges                    = 0x55
DW_AT_trampoline                = 0x56
DW_AT_call_column               = 0x57
DW_AT_call_file                 = 0x58
DW_AT_call_line                 = 0x59
DW_AT_description               = 0x5a
DW_AT_binary_scale              = 0x5b
DW_AT_decimal_scale             = 0x5c
DW_AT_small                     = 0x5d
DW_AT_decimal_sign              = 0x5e
DW_AT_digit_count               = 0x5f
DW_AT_picture_string            = 0x60
DW_AT_mutable                   = 0x61
DW_AT_threads_scaled            = 0x62
DW_AT_explicit                  = 0x63
DW_AT_object_pointer            = 0x64
DW_AT_endianity                 = 0x65
DW_AT_elemental                 = 0x66
DW_AT_pure                      = 0x67
DW_AT_recursive                 = 0x68
DW_AT_signature                 = 0x69
DW_AT_main_subprogram           = 0x6a
DW_AT_data_bit_offset           = 0x6b
DW_AT_const_expr                = 0x6c
DW_AT_enum_class                = 0x6d
DW_AT_linkage_name              = 0x6e
DW_AT_lo_user                   = 0x2000
DW_AT_hi_user                   = 0x3fff

ATTR_MAP = {
    DW_AT_sibling                   : "DW_AT_sibling",
    DW_AT_location                  : "DW_AT_location",
    DW_AT_name                      : "DW_AT_name",
    DW_AT_ordering                  : "DW_AT_ordering",
    DW_AT_byte_size                 : "DW_AT_byte_size",
    DW_AT_bit_offset                : "DW_AT_bit_offset",
    DW_AT_bit_size                  : "DW_AT_bit_size",
    DW_AT_stmt_list                 : "DW_AT_stmt_list",
    DW_AT_low_pc                    : "DW_AT_low_pc",
    DW_AT_high_pc                   : "DW_AT_high_pc",
    DW_AT_language                  : "DW_AT_language",
    DW_AT_discr                     : "DW_AT_discr",
    DW_AT_discr_value               : "DW_AT_discr_value",
    DW_AT_visibility                : "DW_AT_visibility",
    DW_AT_import                    : "DW_AT_import",
    DW_AT_string_length             : "DW_AT_string_length",
    DW_AT_common_reference          : "DW_AT_common_reference",
    DW_AT_comp_dir                  : "DW_AT_comp_dir",
    DW_AT_const_value               : "DW_AT_const_value",
    DW_AT_containing_type           : "DW_AT_containing_type",
    DW_AT_default_value             : "DW_AT_default_value",
    DW_AT_inline                    : "DW_AT_inline",
    DW_AT_is_optional               : "DW_AT_is_optional",
    DW_AT_lower_bound               : "DW_AT_lower_bound",
    DW_AT_producer                  : "DW_AT_producer",
    DW_AT_prototyped                : "DW_AT_prototyped",
    DW_AT_return_addr               : "DW_AT_return_addr",
    DW_AT_start_scope               : "DW_AT_start_scope",
    DW_AT_bit_stride                : "DW_AT_bit_stride",
    DW_AT_upper_bound               : "DW_AT_upper_bound",
    DW_AT_abstract_origin           : "DW_AT_abstract_origin",
    DW_AT_accessibility             : "DW_AT_accessibility",
    DW_AT_address_class             : "DW_AT_address_class",
    DW_AT_artificial                : "DW_AT_artificial",
    DW_AT_base_types                : "DW_AT_base_types",
    DW_AT_calling_convention        : "DW_AT_calling_convention",
    DW_AT_count                     : "DW_AT_count",
    DW_AT_data_member_location      : "DW_AT_data_member_location",
    DW_AT_decl_column               : "DW_AT_decl_column",
    DW_AT_decl_file                 : "DW_AT_decl_file",
    DW_AT_decl_line                 : "DW_AT_decl_line",
    DW_AT_declaration               : "DW_AT_declaration",
    DW_AT_discr_list                : "DW_AT_discr_list",
    DW_AT_encoding                  : "DW_AT_encoding",
    DW_AT_external                  : "DW_AT_external",
    DW_AT_frame_base                : "DW_AT_frame_base",
    DW_AT_friend                    : "DW_AT_friend",
    DW_AT_identifier_case           : "DW_AT_identifier_case",
    DW_AT_macro_info                : "DW_AT_macro_info",
    DW_AT_namelist_item             : "DW_AT_namelist_item",
    DW_AT_priority                  : "DW_AT_priority",
    DW_AT_segment                   : "DW_AT_segment",
    DW_AT_specification             : "DW_AT_specification",
    DW_AT_static_link               : "DW_AT_static_link",
    DW_AT_type                      : "DW_AT_type",
    DW_AT_use_location              : "DW_AT_use_location",
    DW_AT_variable_parameter        : "DW_AT_variable_parameter",
    DW_AT_virtuality                : "DW_AT_virtuality",
    DW_AT_vtable_elem_location      : "DW_AT_vtable_elem_location",
    DW_AT_allocated                 : "DW_AT_allocated",
    DW_AT_associated                : "DW_AT_associated",
    DW_AT_data_location             : "DW_AT_data_location",
    DW_AT_byte_stride               : "DW_AT_byte_stride",
    DW_AT_entry_pc                  : "DW_AT_entry_pc",
    DW_AT_use_UTF8                  : "DW_AT_use_UTF8",
    DW_AT_extension                 : "DW_AT_extension",
    DW_AT_ranges                    : "DW_AT_ranges",
    DW_AT_trampoline                : "DW_AT_trampoline",
    DW_AT_call_column               : "DW_AT_call_column",
    DW_AT_call_file                 : "DW_AT_call_file",
    DW_AT_call_line                 : "DW_AT_call_line",
    DW_AT_description               : "DW_AT_description",
    DW_AT_binary_scale              : "DW_AT_binary_scale",
    DW_AT_decimal_scale             : "DW_AT_decimal_scale",
    DW_AT_small                     : "DW_AT_small",
    DW_AT_decimal_sign              : "DW_AT_decimal_sign",
    DW_AT_digit_count               : "DW_AT_digit_count",
    DW_AT_picture_string            : "DW_AT_picture_string",
    DW_AT_mutable                   : "DW_AT_mutable",
    DW_AT_threads_scaled            : "DW_AT_threads_scaled",
    DW_AT_explicit                  : "DW_AT_explicit",
    DW_AT_object_pointer            : "DW_AT_object_pointer",
    DW_AT_endianity                 : "DW_AT_endianity",
    DW_AT_elemental                 : "DW_AT_elemental",
    DW_AT_pure                      : "DW_AT_pure",
    DW_AT_recursive                 : "DW_AT_recursive",
    DW_AT_signature                 : "DW_AT_signature",
    DW_AT_main_subprogram           : "DW_AT_main_subprogram",
    DW_AT_data_bit_offset           : "DW_AT_data_bit_offset",
    DW_AT_const_expr                : "DW_AT_const_expr",
    DW_AT_enum_class                : "DW_AT_enum_class",
    DW_AT_linkage_name              : "DW_AT_linkage_name",
    DW_AT_lo_user                   : "DW_AT_lo_user",
    DW_AT_hi_user                   : "DW_AT_hi_user",
}

class AttributeEncoding(Base):
    MAP = ATTR_MAP

'''
FORM_MAP = {
    DW_FORM_addr                : "DW_FORM_addr",
    DW_FORM_block2              : "DW_FORM_block2",
    DW_FORM_block4              : "DW_FORM_block4",
    DW_FORM_data2               : "DW_FORM_data2",
    DW_FORM_data4               : "DW_FORM_data4",
    DW_FORM_data8               : "DW_FORM_data8",
    DW_FORM_string              : "DW_FORM_string",
    DW_FORM_block               : "DW_FORM_block",
    DW_FORM_block1              : "DW_FORM_block1",
    DW_FORM_data1               : "DW_FORM_data1",
    DW_FORM_flag                : "DW_FORM_flag",
    DW_FORM_sdata               : "DW_FORM_sdata",
    DW_FORM_strp                : "DW_FORM_strp",
    DW_FORM_udata               : "DW_FORM_udata",
    DW_FORM_ref_addr            : "DW_FORM_ref_addr",
    DW_FORM_ref1                : "DW_FORM_ref1",
    DW_FORM_ref2                : "DW_FORM_ref2",
    DW_FORM_ref4                : "DW_FORM_ref4",
    DW_FORM_ref8                : "DW_FORM_ref8",
    DW_FORM_ref_udata           : "DW_FORM_ref_udata",
    DW_FORM_indirect            : "DW_FORM_indirect",
    DW_FORM_sec_offset          : "DW_FORM_sec_offset",
    DW_FORM_exprloc             : "DW_FORM_exprloc",
    DW_FORM_flag_present        : "DW_FORM_flag_present",
    DW_FORM_ref_sig8            : "DW_FORM_ref_sig8",
}
'''

class AttributeForm(enum.IntEnum):
    #MAP = FORM_MAP
    ##
    ## Attribute form encodings.
    ##
    DW_FORM_addr                    = 0x01
    DW_FORM_block2                  = 0x03
    DW_FORM_block4                  = 0x04
    DW_FORM_data2                   = 0x05
    DW_FORM_data4                   = 0x06
    DW_FORM_data8                   = 0x07
    DW_FORM_string                  = 0x08
    DW_FORM_block                   = 0x09
    DW_FORM_block1                  = 0x0a
    DW_FORM_data1                   = 0x0b
    DW_FORM_flag                    = 0x0c
    DW_FORM_sdata                   = 0x0d
    DW_FORM_strp                    = 0x0e
    DW_FORM_udata                   = 0x0f
    DW_FORM_ref_addr                = 0x10
    DW_FORM_ref1                    = 0x11
    DW_FORM_ref2                    = 0x12
    DW_FORM_ref4                    = 0x13
    DW_FORM_ref8                    = 0x14
    DW_FORM_ref_udata               = 0x15
    DW_FORM_indirect                = 0x16
    DW_FORM_sec_offset              = 0x17
    DW_FORM_exprloc                 = 0x18
    DW_FORM_flag_present            = 0x19
    DW_FORM_ref_sig8                = 0x20

##
## DWARF operation encodings.
##
DW_OP_addr                      = 0x03
DW_OP_deref                     = 0x06
DW_OP_const1u                   = 0x08
DW_OP_const1s                   = 0x09
DW_OP_const2u                   = 0x0a
DW_OP_const2s                   = 0x0b
DW_OP_const4u                   = 0x0c
DW_OP_const4s                   = 0x0d
DW_OP_const8u                   = 0x0e
DW_OP_const8s                   = 0x0f
DW_OP_constu                    = 0x10
DW_OP_consts                    = 0x11
DW_OP_dup                       = 0x12
DW_OP_drop                      = 0x13
DW_OP_over                      = 0x14
DW_OP_pick                      = 0x15
DW_OP_swap                      = 0x16
DW_OP_rot                       = 0x17
DW_OP_xderef                    = 0x18
DW_OP_abs                       = 0x19
DW_OP_and                       = 0x1a
DW_OP_div                       = 0x1b
DW_OP_minus                     = 0x1c
DW_OP_mod                       = 0x1d
DW_OP_mul                       = 0x1e
DW_OP_neg                       = 0x1f
DW_OP_not                       = 0x20
DW_OP_or                        = 0x21
DW_OP_plus                      = 0x22
DW_OP_plus_uconst               = 0x23
DW_OP_shl                       = 0x24
DW_OP_shr                       = 0x25
DW_OP_shra                      = 0x26
DW_OP_xor                       = 0x27
DW_OP_skip                      = 0x2f
DW_OP_bra                       = 0x28
DW_OP_eq                        = 0x29
DW_OP_ge                        = 0x2a
DW_OP_gt                        = 0x2b
DW_OP_le                        = 0x2c
DW_OP_lt                        = 0x2d
DW_OP_ne                        = 0x2e
DW_OP_lit0                      = 0x30
DW_OP_lit1                      = 0x31
DW_OP_lit2                      = 0x32
DW_OP_lit3                      = 0x33
DW_OP_lit4                      = 0x34
DW_OP_lit5                      = 0x35
DW_OP_lit6                      = 0x36
DW_OP_lit7                      = 0x37
DW_OP_lit8                      = 0x38
DW_OP_lit9                      = 0x39
DW_OP_lit10                     = 0x3a
DW_OP_lit11                     = 0x3b
DW_OP_lit12                     = 0x3c
DW_OP_lit13                     = 0x3d
DW_OP_lit14                     = 0x3e
DW_OP_lit15                     = 0x3f
DW_OP_lit16                     = 0x40
DW_OP_lit17                     = 0x41
DW_OP_lit18                     = 0x42
DW_OP_lit19                     = 0x43
DW_OP_lit20                     = 0x44
DW_OP_lit21                     = 0x45
DW_OP_lit22                     = 0x46
DW_OP_lit23                     = 0x47
DW_OP_lit24                     = 0x48
DW_OP_lit25                     = 0x49
DW_OP_lit26                     = 0x4a
DW_OP_lit27                     = 0x4b
DW_OP_lit28                     = 0x4c
DW_OP_lit29                     = 0x4d
DW_OP_lit30                     = 0x4e
DW_OP_lit31                     = 0x4f
DW_OP_reg0                      = 0x50
DW_OP_reg1                      = 0x51
DW_OP_reg2                      = 0x52
DW_OP_reg3                      = 0x53
DW_OP_reg4                      = 0x54
DW_OP_reg5                      = 0x55
DW_OP_reg6                      = 0x56
DW_OP_reg7                      = 0x57
DW_OP_reg8                      = 0x58
DW_OP_reg9                      = 0x59
DW_OP_reg10                     = 0x5a
DW_OP_reg11                     = 0x5b
DW_OP_reg12                     = 0x5c
DW_OP_reg13                     = 0x5d
DW_OP_reg14                     = 0x5e
DW_OP_reg15                     = 0x5f
DW_OP_reg16                     = 0x60
DW_OP_reg17                     = 0x61
DW_OP_reg18                     = 0x62
DW_OP_reg19                     = 0x63
DW_OP_reg20                     = 0x64
DW_OP_reg21                     = 0x65
DW_OP_reg22                     = 0x66
DW_OP_reg23                     = 0x67
DW_OP_reg24                     = 0x68
DW_OP_reg25                     = 0x69
DW_OP_reg26                     = 0x6a
DW_OP_reg27                     = 0x6b
DW_OP_reg28                     = 0x6c
DW_OP_reg29                     = 0x6d
DW_OP_reg30                     = 0x6e
DW_OP_reg31                     = 0x6f
DW_OP_breg0                     = 0x70
DW_OP_breg1                     = 0x71
DW_OP_breg2                     = 0x72
DW_OP_breg3                     = 0x73
DW_OP_breg4                     = 0x74
DW_OP_breg5                     = 0x75
DW_OP_breg6                     = 0x76
DW_OP_breg7                     = 0x77
DW_OP_breg8                     = 0x78
DW_OP_breg9                     = 0x79
DW_OP_breg10                    = 0x7a
DW_OP_breg11                    = 0x7b
DW_OP_breg12                    = 0x7c
DW_OP_breg13                    = 0x7d
DW_OP_breg14                    = 0x7e
DW_OP_breg15                    = 0x7f
DW_OP_breg16                    = 0x80
DW_OP_breg17                    = 0x81
DW_OP_breg18                    = 0x82
DW_OP_breg19                    = 0x83
DW_OP_breg20                    = 0x84
DW_OP_breg21                    = 0x85
DW_OP_breg22                    = 0x86
DW_OP_breg23                    = 0x87
DW_OP_breg24                    = 0x88
DW_OP_breg25                    = 0x89
DW_OP_breg26                    = 0x8a
DW_OP_breg27                    = 0x8b
DW_OP_breg28                    = 0x8c
DW_OP_breg29                    = 0x8d
DW_OP_breg30                    = 0x8e
DW_OP_breg31                    = 0x8f
DW_OP_regx                      = 0x90
DW_OP_fbreg                     = 0x91
DW_OP_bregx                     = 0x92
DW_OP_piece                     = 0x93
DW_OP_deref_size                = 0x94
DW_OP_xderef_size               = 0x95
DW_OP_nop                       = 0x96
DW_OP_push_object_address       = 0x97
DW_OP_call2                     = 0x98
DW_OP_call4                     = 0x99
DW_OP_call_ref                  = 0x9a
DW_OP_form_tls_address          = 0x9b
DW_OP_call_frame_cfa            = 0x9c
DW_OP_bit_piece                 = 0x9d
DW_OP_implicit_value            = 0x9e
DW_OP_stack_value               = 0x9f
DW_OP_lo_user                   = 0xe0
DW_OP_hi_user                   = 0xff

OPERATION_MAP = {
    DW_OP_addr                  : "DW_OP_addr",
    DW_OP_deref                 : "DW_OP_deref",
    DW_OP_const1u               : "DW_OP_const1u",
    DW_OP_const1s               : "DW_OP_const1s",
    DW_OP_const2u               : "DW_OP_const2u",
    DW_OP_const2s               : "DW_OP_const2s",
    DW_OP_const4u               : "DW_OP_const4u",
    DW_OP_const4s               : "DW_OP_const4s",
    DW_OP_const8u               : "DW_OP_const8u",
    DW_OP_const8s               : "DW_OP_const8s",
    DW_OP_constu                : "DW_OP_constu",
    DW_OP_consts                : "DW_OP_consts",
    DW_OP_dup                   : "DW_OP_dup",
    DW_OP_drop                  : "DW_OP_drop",
    DW_OP_over                  : "DW_OP_over",
    DW_OP_pick                  : "DW_OP_pick",
    DW_OP_swap                  : "DW_OP_swap",
    DW_OP_rot                   : "DW_OP_rot",
    DW_OP_xderef                : "DW_OP_xderef",
    DW_OP_abs                   : "DW_OP_abs",
    DW_OP_and                   : "DW_OP_and",
    DW_OP_div                   : "DW_OP_div",
    DW_OP_minus                 : "DW_OP_minus",
    DW_OP_mod                   : "DW_OP_mod",
    DW_OP_mul                   : "DW_OP_mul",
    DW_OP_neg                   : "DW_OP_neg",
    DW_OP_not                   : "DW_OP_not",
    DW_OP_or                    : "DW_OP_or",
    DW_OP_plus                  : "DW_OP_plus",
    DW_OP_plus_uconst           : "DW_OP_plus_uconst",
    DW_OP_shl                   : "DW_OP_shl",
    DW_OP_shr                   : "DW_OP_shr",
    DW_OP_shra                  : "DW_OP_shra",
    DW_OP_xor                   : "DW_OP_xor",
    DW_OP_skip                  : "DW_OP_skip",
    DW_OP_bra                   : "DW_OP_bra",
    DW_OP_eq                    : "DW_OP_eq",
    DW_OP_ge                    : "DW_OP_ge",
    DW_OP_gt                    : "DW_OP_gt",
    DW_OP_le                    : "DW_OP_le",
    DW_OP_lt                    : "DW_OP_lt",
    DW_OP_ne                    : "DW_OP_ne",
    DW_OP_lit0                  : "DW_OP_lit0",
    DW_OP_lit1                  : "DW_OP_lit1",
    DW_OP_lit2                  : "DW_OP_lit2",
    DW_OP_lit3                  : "DW_OP_lit3",
    DW_OP_lit4                  : "DW_OP_lit4",
    DW_OP_lit5                  : "DW_OP_lit5",
    DW_OP_lit6                  : "DW_OP_lit6",
    DW_OP_lit7                  : "DW_OP_lit7",
    DW_OP_lit8                  : "DW_OP_lit8",
    DW_OP_lit9                  : "DW_OP_lit9",
    DW_OP_lit10                 : "DW_OP_lit10",
    DW_OP_lit11                 : "DW_OP_lit11",
    DW_OP_lit12                 : "DW_OP_lit12",
    DW_OP_lit13                 : "DW_OP_lit13",
    DW_OP_lit14                 : "DW_OP_lit14",
    DW_OP_lit15                 : "DW_OP_lit15",
    DW_OP_lit16                 : "DW_OP_lit16",
    DW_OP_lit17                 : "DW_OP_lit17",
    DW_OP_lit18                 : "DW_OP_lit18",
    DW_OP_lit19                 : "DW_OP_lit19",
    DW_OP_lit20                 : "DW_OP_lit20",
    DW_OP_lit21                 : "DW_OP_lit21",
    DW_OP_lit22                 : "DW_OP_lit22",
    DW_OP_lit23                 : "DW_OP_lit23",
    DW_OP_lit24                 : "DW_OP_lit24",
    DW_OP_lit25                 : "DW_OP_lit25",
    DW_OP_lit26                 : "DW_OP_lit26",
    DW_OP_lit27                 : "DW_OP_lit27",
    DW_OP_lit28                 : "DW_OP_lit28",
    DW_OP_lit29                 : "DW_OP_lit29",
    DW_OP_lit30                 : "DW_OP_lit30",
    DW_OP_lit31                 : "DW_OP_lit31",
    DW_OP_reg0                  : "DW_OP_reg0",
    DW_OP_reg1                  : "DW_OP_reg1",
    DW_OP_reg2                  : "DW_OP_reg2",
    DW_OP_reg3                  : "DW_OP_reg3",
    DW_OP_reg4                  : "DW_OP_reg4",
    DW_OP_reg5                  : "DW_OP_reg5",
    DW_OP_reg6                  : "DW_OP_reg6",
    DW_OP_reg7                  : "DW_OP_reg7",
    DW_OP_reg8                  : "DW_OP_reg8",
    DW_OP_reg9                  : "DW_OP_reg9",
    DW_OP_reg10                 : "DW_OP_reg10",
    DW_OP_reg11                 : "DW_OP_reg11",
    DW_OP_reg12                 : "DW_OP_reg12",
    DW_OP_reg13                 : "DW_OP_reg13",
    DW_OP_reg14                 : "DW_OP_reg14",
    DW_OP_reg15                 : "DW_OP_reg15",
    DW_OP_reg16                 : "DW_OP_reg16",
    DW_OP_reg17                 : "DW_OP_reg17",
    DW_OP_reg18                 : "DW_OP_reg18",
    DW_OP_reg19                 : "DW_OP_reg19",
    DW_OP_reg20                 : "DW_OP_reg20",
    DW_OP_reg21                 : "DW_OP_reg21",
    DW_OP_reg22                 : "DW_OP_reg22",
    DW_OP_reg23                 : "DW_OP_reg23",
    DW_OP_reg24                 : "DW_OP_reg24",
    DW_OP_reg25                 : "DW_OP_reg25",
    DW_OP_reg26                 : "DW_OP_reg26",
    DW_OP_reg27                 : "DW_OP_reg27",
    DW_OP_reg28                 : "DW_OP_reg28",
    DW_OP_reg29                 : "DW_OP_reg29",
    DW_OP_reg30                 : "DW_OP_reg30",
    DW_OP_reg31                 : "DW_OP_reg31",
    DW_OP_breg0                 : "DW_OP_breg0",
    DW_OP_breg1                 : "DW_OP_breg1",
    DW_OP_breg2                 : "DW_OP_breg2",
    DW_OP_breg3                 : "DW_OP_breg3",
    DW_OP_breg4                 : "DW_OP_breg4",
    DW_OP_breg5                 : "DW_OP_breg5",
    DW_OP_breg6                 : "DW_OP_breg6",
    DW_OP_breg7                 : "DW_OP_breg7",
    DW_OP_breg8                 : "DW_OP_breg8",
    DW_OP_breg9                 : "DW_OP_breg9",
    DW_OP_breg10                : "DW_OP_breg10",
    DW_OP_breg11                : "DW_OP_breg11",
    DW_OP_breg12                : "DW_OP_breg12",
    DW_OP_breg13                : "DW_OP_breg13",
    DW_OP_breg14                : "DW_OP_breg14",
    DW_OP_breg15                : "DW_OP_breg15",
    DW_OP_breg16                : "DW_OP_breg16",
    DW_OP_breg17                : "DW_OP_breg17",
    DW_OP_breg18                : "DW_OP_breg18",
    DW_OP_breg19                : "DW_OP_breg19",
    DW_OP_breg20                : "DW_OP_breg20",
    DW_OP_breg21                : "DW_OP_breg21",
    DW_OP_breg22                : "DW_OP_breg22",
    DW_OP_breg23                : "DW_OP_breg23",
    DW_OP_breg24                : "DW_OP_breg24",
    DW_OP_breg25                : "DW_OP_breg25",
    DW_OP_breg26                : "DW_OP_breg26",
    DW_OP_breg27                : "DW_OP_breg27",
    DW_OP_breg28                : "DW_OP_breg28",
    DW_OP_breg28                : "DW_OP_breg28",
    DW_OP_breg30                : "DW_OP_breg30",
    DW_OP_breg31                : "DW_OP_breg31",
    DW_OP_regx                  : "DW_OP_regx",
    DW_OP_fbreg                 : "DW_OP_fbreg",
    DW_OP_bregx                 : "DW_OP_bregx",
    DW_OP_piece                 : "DW_OP_piece",
    DW_OP_deref_size            : "DW_OP_deref_size",
    DW_OP_xderef_size           : "DW_OP_xderef_size",
    DW_OP_nop                   : "DW_OP_nop",
    DW_OP_push_object_address   : "DW_OP_push_object_address",
    DW_OP_call2                 : "DW_OP_call2",
    DW_OP_call4                 : "DW_OP_call4",
    DW_OP_call_ref              : "DW_OP_call_ref",
    DW_OP_form_tls_address      : "DW_OP_form_tls_address",
    DW_OP_call_frame_cfa        : "DW_OP_call_frame_cfa",
    DW_OP_bit_piece             : "DW_OP_bit_piece",
    DW_OP_implicit_value        : "DW_OP_implicit_value",
    DW_OP_stack_value           : "DW_OP_stack_value",
    DW_OP_lo_user               : "DW_OP_lo_user",
    DW_OP_hi_user               : "DW_OP_hi_user",
}

class Operation(Base):
    MAP = OPERATION_MAP


##
## Base type encoding values.
##
DW_ATE_address                  = 0x01
DW_ATE_boolean                  = 0x02
DW_ATE_complex_float            = 0x03
DW_ATE_float                    = 0x04
DW_ATE_signed                   = 0x05
DW_ATE_signed_char              = 0x06
DW_ATE_unsigned                 = 0x07
DW_ATE_unsigned_char            = 0x08
DW_ATE_imaginary_float          = 0x09
DW_ATE_packed_decimal           = 0x0a
DW_ATE_numeric_string           = 0x0b
DW_ATE_edited                   = 0x0c
DW_ATE_signed_fixed             = 0x0d
DW_ATE_unsigned_fixed           = 0x0e
DW_ATE_decimal_float            = 0x0f
DW_ATE_UTF                      = 0x10
DW_ATE_lo_user                  = 0x80
DW_ATE_hi_user                  = 0xff

ATE_MAP = {
    DW_ATE_address              : "DW_ATE_address",
    DW_ATE_boolean              : "DW_ATE_boolean",
    DW_ATE_complex_float        : "DW_ATE_complex_float",
    DW_ATE_float                : "DW_ATE_float",
    DW_ATE_signed               : "DW_ATE_signed",
    DW_ATE_signed_char          : "DW_ATE_signed_char",
    DW_ATE_unsigned             : "DW_ATE_unsigned",
    DW_ATE_unsigned_char        : "DW_ATE_unsigned_char",
    DW_ATE_imaginary_float      : "DW_ATE_imaginary_float",
    DW_ATE_packed_decimal       : "DW_ATE_packed_decimal",
    DW_ATE_numeric_string       : "DW_ATE_numeric_string",
    DW_ATE_edited               : "DW_ATE_edited",
    DW_ATE_signed_fixed         : "DW_ATE_signed_fixed",
    DW_ATE_unsigned_fixed       : "DW_ATE_unsigned_fixed",
    DW_ATE_decimal_float        : "DW_ATE_decimal_float",
    DW_ATE_UTF                  : "DW_ATE_UTF",
    DW_ATE_lo_user              : "DW_ATE_lo_user",
    DW_ATE_hi_user              : "DW_ATE_hi_user",
}

class BaseTypeEncoding(Base):
    MAP = ATE_MAP


##
## Decimal sign encodings.
##
DW_DS_unsigned                  = 0x01
DW_DS_leading_overpunch         = 0x02
DW_DS_trailing_overpunch        = 0x03
DW_DS_leading_separate          = 0x04
DW_DS_trailing_separate         = 0x05

DS_MAP = {
    DW_DS_unsigned              : "DW_DS_unsigned",
    DW_DS_leading_overpunch     : "DW_DS_leading_overpunch",
    DW_DS_trailing_overpunch    : "DW_DS_trailing_overpunch",
    DW_DS_leading_separate      : "DW_DS_leading_separate",
    DW_DS_trailing_separate     : "DW_DS_trailing_separate",
}

class DecimalSign(Base):
    MAP = DS_MAP

##
## Endianity encoding.
##
DW_END_default                  = 0x00
DW_END_big                      = 0x01
DW_END_little                   = 0x02
DW_END_lo_user                  = 0x40
DW_END_hi_user                  = 0xff

END_MAP = {
    DW_END_default              : "DW_END_default",
    DW_END_big                  : "DW_END_big",
    DW_END_little               : "DW_END_little",
    DW_END_lo_user              : "DW_END_lo_user",
    DW_END_hi_user              : "DW_END_hi_user",
}

class Endianity(Base):
    MAP = END_MAP

##
## Accessibility encodings.
##
DW_ACCESS_public                = 0x01
DW_ACCESS_protected             = 0x02
DW_ACCESS_private               = 0x03

ACCESS_MAP = {
    DW_ACCESS_public            : "DW_ACCESS_public",
    DW_ACCESS_protected         : "DW_ACCESS_protected",
    DW_ACCESS_private           : "DW_ACCESS_private",
}

class Accessibility(Base):
    MAP = ACCESS_MAP

##
## Visibility encodings.
##
DW_VIS_local                    = 0x01
DW_VIS_exported                 = 0x02
DW_VIS_qualified                = 0x03

VIS_MAP = {
    DW_VIS_local                : "DW_VIS_local",
    DW_VIS_exported             : "DW_VIS_exported",
    DW_VIS_qualified            : "DW_VIS_qualified",
}

class Visibility(Base):
    MAP = VIS_MAP

##
## Virtuality encodings.
##
DW_VIRTUALITY_none              = 0x00
DW_VIRTUALITY_virtual           = 0x01
DW_VIRTUALITY_pure_virtual      = 0x02

VIRTUALITY_MAP = {
    DW_VIRTUALITY_none          : "DW_VIRTUALITY_none",
    DW_VIRTUALITY_virtual       : "DW_VIRTUALITY_virtual",
    DW_VIRTUALITY_pure_virtual  : "DW_VIRTUALITY_pure_virtual",
}

class Virtuality(Base):
    MAP = VIRTUALITY_MAP


##
## Language encodings.
##
DW_LANG_C89                     = 0x0001    # 0
DW_LANG_C                       = 0x0002    # 0
DW_LANG_Ada83                   = 0x0003    # 1
DW_LANG_C_plus_plus             = 0x0004    # 0
DW_LANG_Cobol74                 = 0x0005    # 1
DW_LANG_Cobol85                 = 0x0006    # 1
DW_LANG_Fortran77               = 0x0007    # 1
DW_LANG_Fortran90               = 0x0008    # 1
DW_LANG_Pascal83                = 0x0009    # 1
DW_LANG_Modula2                 = 0x000a    # 1
DW_LANG_Java                    = 0x000b    # 0
DW_LANG_C99                     = 0x000c    # 0
DW_LANG_Ada95                   = 0x000d    # 1
DW_LANG_Fortran95               = 0x000e    # 1
DW_LANG_PLI                     = 0x000f    # 1
DW_LANG_ObjC                    = 0x0010    # 0
DW_LANG_ObjC_plus_plus          = 0x0011    # 0
DW_LANG_UPC                     = 0x0012    # 0
DW_LANG_D                       = 0x0013    # 0
DW_LANG_Python                  = 0x0014    # 0
DW_LANG_lo_user                 = 0x8000
DW_LANG_hi_user                 = 0xffff

LANG_MAP = {
    DW_LANG_C89                 : "DW_LANG_C89",
    DW_LANG_C                   : "DW_LANG_C",
    DW_LANG_Ada83               : "DW_LANG_Ada83",
    DW_LANG_C_plus_plus         : "DW_LANG_C_plus_plus",
    DW_LANG_Cobol74             : "DW_LANG_Cobol74",
    DW_LANG_Cobol85             : "DW_LANG_Cobol85",
    DW_LANG_Fortran77           : "DW_LANG_Fortran77",
    DW_LANG_Fortran90           : "DW_LANG_Fortran90",
    DW_LANG_Pascal83            : "DW_LANG_Pascal83",
    DW_LANG_Modula2             : "DW_LANG_Modula2",
    DW_LANG_Java                : "DW_LANG_Java",
    DW_LANG_C99                 : "DW_LANG_C99",
    DW_LANG_Ada95               : "DW_LANG_Ada95",
    DW_LANG_Fortran95           : "DW_LANG_Fortran95",
    DW_LANG_PLI                 : "DW_LANG_PLI",
    DW_LANG_ObjC                : "DW_LANG_ObjC",
    DW_LANG_ObjC_plus_plus      : "DW_LANG_ObjC_plus_plus",
    DW_LANG_UPC                 : "DW_LANG_UPC",
    DW_LANG_D                   : "DW_LANG_D",
    DW_LANG_Python              : "DW_LANG_Python",
    DW_LANG_lo_user             : "DW_LANG_lo_user",
    DW_LANG_hi_user             : "DW_LANG_hi_user",
}

class Language(Base):
    MAP = LANG_MAP

##
## Identifier case encodings.
##
DW_ID_case_sensitive            = 0x00
DW_ID_up_case                   = 0x01
DW_ID_down_case                 = 0x02
DW_ID_case_insensitive          = 0x03

ID_MAP = {
    DW_ID_case_sensitive        : "DW_ID_case_sensitive",
    DW_ID_up_case               : "DW_ID_up_case",
    DW_ID_down_case             : "DW_ID_down_case",
    DW_ID_case_insensitive      : "DW_ID_case_insensitive",
}

class IdentifierCase(Base):
    MAP = ID_MAP

##
## Calling convention encodings.
##
DW_CC_normal                    = 0x01
DW_CC_program                   = 0x02
DW_CC_nocall                    = 0x03
DW_CC_lo_user                   = 0x40
DW_CC_hi_user                   = 0xff

CC_MAP = {
    DW_CC_normal                : "DW_CC_normal",
    DW_CC_program               : "DW_CC_program",
    DW_CC_nocall                : "DW_CC_nocall",
    DW_CC_lo_user               : "DW_CC_lo_user",
    DW_CC_hi_user               : "DW_CC_hi_user",
}

class CallingConvention(Base):
    MAP = CC_MAP

##
## Inline encodings.
##
DW_INL_not_inlined              = 0x00
DW_INL_inlined                  = 0x01
DW_INL_declared_not_inlined     = 0x02
DW_INL_declared_inlined         = 0x03

INL_MAP = {
    DW_INL_not_inlined          : "DW_INL_not_inlined",
    DW_INL_inlined              : "DW_INL_inlined",
    DW_INL_declared_not_inlined : "DW_INL_declared_not_inlined",
    DW_INL_declared_inlined     : "DW_INL_declared_inlined",
}

class Inline(Base):
    MAP = INL_MAP


##
## Ordering encodings.
##
DW_ORD_row_major                = 0x00
DW_ORD_col_major                = 0x01

ORD_MAP = {
    DW_ORD_row_major            : "DW_ORD_row_major",
    DW_ORD_col_major            : "DW_ORD_col_major",
}

class Ordering(Base):
    MAP = ORD_MAP


##
##  Discriminant descriptor encodings.
##
DW_DSC_label                    = 0x00
DW_DSC_range                    = 0x01

DSC_MAP = {
    DW_DSC_label                : "DW_DSC_label",
    DW_DSC_range                : "DW_DSC_range",
}

class DiscriminantDescriptor(Base):
    MAP = DSC_MAP


##
##  Line Number Standard Opcode Encodings.
##
DW_LNS_copy                     = 0x01
DW_LNS_advance_pc               = 0x02
DW_LNS_advance_line             = 0x03
DW_LNS_set_file                 = 0x04
DW_LNS_set_column               = 0x05
DW_LNS_negate_stmt              = 0x06
DW_LNS_set_basic_block          = 0x07
DW_LNS_const_add_pc             = 0x08
DW_LNS_fixed_advance_pc         = 0x09
DW_LNS_set_prologue_end         = 0x0a
DW_LNS_set_epilogue_begin       = 0x0b
DW_LNS_set_isa                  = 0x0c

LNS_MAP = {
    DW_LNS_copy                 : "DW_LNS_copy",
    DW_LNS_advance_pc           : "DW_LNS_advance_pc",
    DW_LNS_advance_line         : "DW_LNS_advance_line",
    DW_LNS_set_file             : "DW_LNS_set_file",
    DW_LNS_set_column           : "DW_LNS_set_column",
    DW_LNS_negate_stmt          : "DW_LNS_negate_stmt",
    DW_LNS_set_basic_block      : "DW_LNS_set_basic_block",
    DW_LNS_const_add_pc         : "DW_LNS_const_add_pc",
    DW_LNS_fixed_advance_pc     : "DW_LNS_fixed_advance_pc",
    DW_LNS_set_prologue_end     : "DW_LNS_set_prologue_end",
    DW_LNS_set_epilogue_begin   : "DW_LNS_set_epilogue_begin",
    DW_LNS_set_isa              : "DW_LNS_set_isa",
}

class LineNumberStandard(Base):
    MAP = LNS_MAP


##
##  Line Number Extended Opcode Encodings.
##
DW_LNE_end_sequence             = 0x01
DW_LNE_set_address              = 0x02
DW_LNE_define_file              = 0x03
DW_LNE_set_discriminator        = 0x04
DW_LNE_lo_user                  = 0x80
DW_LNE_hi_user                  = 0xff

LNE_MAP = {
    DW_LNE_end_sequence         : "DW_LNE_end_sequence",
    DW_LNE_set_address          : "DW_LNE_set_address",
    DW_LNE_define_file          : "DW_LNE_define_file",
    DW_LNE_set_discriminator    : "DW_LNE_set_discriminator",
    DW_LNE_lo_user              : "DW_LNE_lo_user",
    DW_LNE_hi_user              : "DW_LNE_hi_user",
}


class LineNumberExtended(Base):
    MAP = LNE_MAP

##
## Macro Information.
##
DW_MACINFO_define               = 0x01
DW_MACINFO_undef                = 0x02
DW_MACINFO_start_file           = 0x03
DW_MACINFO_end_file             = 0x04
DW_MACINFO_vendor_ext           = 0xff

MACINFO_MAP = {
    DW_MACINFO_define           : "DW_MACINFO_define",
    DW_MACINFO_undef            : "DW_MACINFO_undef",
    DW_MACINFO_start_file       : "DW_MACINFO_start_file",
    DW_MACINFO_end_file         : "DW_MACINFO_end_file",
    DW_MACINFO_vendor_ext       : "DW_MACINFO_vendor_ext",
}

class MacroInformation(Base):
    MAP = MACINFO_MAP

##
## Segmented Addresses.
##

# For example, the Intel386 processor might use the following values:

DW_ADDR_none                    = 0x00
DW_ADDR_near16                  = 0x01
DW_ADDR_far16                   = 0x02
DW_ADDR_huge16                  = 0x03
DW_ADDR_near32                  = 0x04
DW_ADDR_far32                   = 0x05


class ExceptionHeaderEncoding(enum.IntEnum):
    """The DWARF Exception Header Encoding is used to describe the type of data
    used in the .eh_frame and .eh_frame_hdr section.

    Linux specific, s. Linux Base Standard.
    """
    DW_EH_PE_absptr     = 0x00
    DW_EH_PE_uleb128    = 0x01
    DW_EH_PE_udata2     = 0x02
    DW_EH_PE_udata4     = 0x03
    DW_EH_PE_udata8     = 0x04
    DW_EH_PE_sleb128    = 0x09
    DW_EH_PE_sdata2     = 0x0A
    DW_EH_PE_sdata4     = 0x0B
    DW_EH_PE_sdata8     = 0x0C
    DW_EH_PE_pcrel      = 0x10
    DW_EH_PE_textrel    = 0x20
    DW_EH_PE_datarel    = 0x30
    DW_EH_PE_funcrel    = 0x40
    DW_EH_PE_aligned    = 0x50


class CFIExtensions(enum.IntEnum):
    """
    """

    DW_CFA_GNU_args_size                = 0x2e
    DW_CFA_GNU_negative_offset_extended = 0x2f



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import division

"""Parser for ELF files.
"""

__copyright__ = """
   objutils - Object file library for Python.

  (C) 2010-2020 by Christoph Schueler <cpu12.gems@googlemail.com>

  All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import binascii
from collections import namedtuple, OrderedDict
import hashlib
from itertools import groupby
import re
import time


from construct import Adapter, Array, CString, Const, Enum, IfThenElse, Padding, Struct
from construct import BitStruct, BitsInteger, Bytes, Construct, this
from construct import Computed, Pass, Tell, Union, singleton
from construct import Int8ul, Int16ul, Int32ul, Int32sl, Int64ul, Int64sl
from construct import         Int16ub, Int32ub, Int32sb, Int64ub, Int64sb

from sqlalchemy import (func, not_
)

from objutils import Image, Section
from objutils.utils import create_memorymapped_fileview
from objutils.elf import defs, model
from objutils.elf.arm import attributes


MAGIC = b'\x7fELF'


@singleton
class Pass2(Construct):
    r"""
    A do-nothing construct, useful as the default case for Switch. Returns None on parsing, puts nothing on building.
    Example::
        >>> Pass.parse(b"")
        >>> Pass.build(None)
        b''
        >>> Pass.sizeof()
        0
    """
    def __init__(self):
        super().__init__()
        self.flagbuildnone = True
    def _parse(self, stream, context, path):
        return None
    def _build(self, obj, stream, context, path):
        pass
    def _sizeof(self, context, path):
        return 0

Test = Struct(
    "ei_version"  / Enum(Int8ul, default = Pass,
        EV_NONE = 0,
        EV_CURRENT = 1,
    )
)


class UTCTimeStampAdapter(Adapter):

    def _decode(self, obj, context):
        return time.ctime(obj)

    def _encode(self, obj, context):
        return int(time.mktime(time.strptime(obj)))


def UTCTimeStamp(name):
    return UTCTimeStampAdapter(name / Int32ul)


class ListToBytesAdapter(Adapter):

    def _decode(self, obj, context):
        return bytes(obj)

    def _encode(self, obj, context):
        return list(obj)


class PrintContext(Construct):

    def _parse(self, stream, context, *args, **kws):
        print("CTX: {} {} {}".format(context, args, kws))
        print("CTX_END")


class ElfFileStruct(Struct):

    def __init__(self, *args, **kws):
        super(ElfFileStruct, self).__init__(*args, **kws)

    def elf32Addr(self):
        pass

    def setByteOrder(self):
        print("*** {}".format(self))


DebugInfo = namedtuple('DebugInfo', 'section image')


class DBAPI:
    """

    """

    def __init__(self, parent):
        self._session = parent.session

    @property
    def session(self):
        return self._session

    @property
    def query(self):
        return self.session.query


class SectionAPI(DBAPI):
    """
    """

    def get(self, section_name: str):
        """Get a single section.

        Parameters
        ----------

        section_name: str

        Returns
        -------
        """

        query = self.query(model.Elf_Section)
        query = query.filter(model.Elf_Section.section_name == section_name)
        return query.first()

    def fetch(self, sections: str = None, name_pattern: str = None, order_by_address: bool = True):
        """

        Returns
        -------
        """

        query = self.query(model.Elf_Section)
        if name_pattern:
            query = query.filter(func.regexp(model.Elf_Section.section_name, name_pattern))
        if order_by_address:
            query = query.order_by(model.Elf_Section.sh_addr)
        else:
            query = query.order_by(model.Elf_Section.section_name)
        result = query.all()
        return result


        #query = query.order_by(model.Elf_Symbol.section_name)

        #for key, value in groupby(query.all(), lambda s: s.section_name):
        #    result[key] = list(value)
        #return result


class SymbolAPI(DBAPI):
    """
    """

    def get(self, symbol_name: str, section_name: str = None):
        """Get a single symbol.

        Parameters
        ----------

        symbol_name: str or list of strings.
            Symbol(s) to look for.

        section_name: str
            Consider only named section, else lookup all sections.

        Returns
        -------
        """
        if isinstance(symbol_name, str):
            names = [symbol_name]

        query = self.query(model.Elf_Symbol)
        if section_name:
            query = query.filter(model.Elf_Symbol.section_name == section_name)
        query = query.filter(model.Elf_Symbol.symbol_name == symbol_name)
        return query.first()

    def fetch(self, sections: str = None, name_pattern: str = None,
        bindings : str = None, access: str = None, types_str: str = None, order_by_value: bool = True):
        """

        Returns
        -------
        """
        result = OrderedDict()
        query = self.query(model.Elf_Symbol)
        if sections:
            sections = [s for s in re.split(r"[ ,]", sections) if s]
            query = query.filter(model.Elf_Symbol.section_name.in_(sections))
        if access:
            access = [a for a in re.split(r"[ ,]", access) if a]
            value = 0
            for item in access:
                item = item.lower()
                if item == "a":
                    value |= defs.SectionFlags.SHF_ALLOC
                elif item == "w":
                    value |= defs.SectionFlags.SHF_WRITE
                elif item == "x":
                    value |= defs.SectionFlags.SHF_EXECINSTR
            query = query.filter(model.Elf_Symbol.access.op('&')(value))
        if bindings:
            bindings = [b for b in re.split(r"[ ,]", bindings) if b]
            flt = []
            for item in bindings:
                item = item.lower()
                if item == "g":
                    flt.append(defs.SymbolBinding.STB_GLOBAL)
                elif item == "l":
                    flt.append(defs.SymbolBinding.STB_LOCAL)
                elif item == "w":
                    flt.append(defs.SymbolBinding.STB_WEAK)
            query = query.filter(model.Elf_Symbol.st_bind.in_(flt))
        if types_str:
            types_arr = [t for t in re.split(r"[ ,]", types_str) if t]
            flt = []
            for item in types_arr:
                item = item.lower()
                if item == "notype":
                    flt.append(defs.SymbolType.STT_NOTYPE)
                elif item == "object":
                    flt.append(defs.SymbolType.STT_OBJECT)
                elif item == "func":
                    flt.append(defs.SymbolType.STT_FUNC)
                elif item == "section":
                    flt.append(defs.SymbolType.STT_SECTION)
                elif item == "file":
                    flt.append(defs.SymbolType.STT_FILE)
                elif item == "common":
                    flt.append(defs.SymbolType.STT_COMMON)
                elif item == "tls":
                    flt.append(defs.SymbolType.STT_TLS)
                query = query.filter(model.Elf_Symbol.st_type.in_(flt))
        if name_pattern:
            query = query.filter(func.regexp(model.Elf_Symbol.symbol_name, name_pattern))
        query = query.order_by(model.Elf_Symbol.section_name)
        if order_by_value:
            query = query.order_by(model.Elf_Symbol.st_value)
        else:
            query = query.order_by(model.Elf_Symbol.symbol_name)
        for key, value in groupby(query.all(), lambda s: s.section_name):
            result[key] = list(value)
        return result

    def fetch_gcc_special_symbols(self):
        """

        Returns
        -------

        Note
        ----
        This is a GCC specific feature and will not work with other compilers.
        """
        syms = self.query(model.Elf_Symbol).filter(model.Elf_Symbol.st_shndx == defs.SectionName.SHN_ABS)
        syms = syms.filter(model.Elf_Symbol.symbol_name != "")
        syms = syms.filter(func.regexp(model.Elf_Symbol.symbol_name, "_.*"))
        syms = syms.filter(model.Elf_Symbol.st_type == defs.SymbolType.STT_NOTYPE)
        return syms.order_by(model.Elf_Symbol.st_value).all()


def calculate_crypto_hash(data):
    sha = hashlib.sha512(data)
    return sha.hexdigest()


class ElfParser(object):
    """
    """

    EI_NIDENT = 16

    DATATYPES32 = {
        "Addr":   (Int32ul, Int32ub),   # 4 - Unsigned program address
        "Half":   (Int16ul, Int16ub),   # 2 - Unsigned medium integer
        "Off":    (Int32ul, Int32ub),   # 4 - Unsigned file offset
        "Sword":  (Int32sl, Int32sb),   # 4 - Signed large integer
        "Word":   (Int32ul, Int32ub),   # 4 - Unsigned large integer
        "Xword":  (Int32ul, Int32ub),   # 8 - Unsigned long integer
        "Sxword": (None, None),         # 8 - Signed long integer
    }

    DATATYPES64 = {
        "Addr":   (Int64ul, Int64ub),   # 8 - Unsigned program address
        "Off":    (Int64ul, Int64ub),   # 8 - Unsigned file offset
        "Half":   (Int16ul, Int16ub),   # 2 - Unsigned medium integer
        "Word":   (Int32ul, Int32ub),   # 4 - Unsigned integer
        "Sword":  (Int32sl, Int32sb),   # 4 - Signed integer
        "Xword":  (Int64ul, Int64ub),   # 8 - Unsigned long integer
        "Sxword": (Int64sl, Int64sb),   # 8 - Signed long integer
    }

    BasicHeader = Struct(
        "header" / Union(None,
            "fields"/ Struct(
                Const(MAGIC),
                "ei_class"  / Int8ul,
                "ei_data"  / Int8ul,
                "ei_version"  / Int8ul,
                "ei_osabi" / Int8ul,
                "ei_abiversion" / Int8ul,
                Padding(7),
            ),
            "bytes" / Bytes(EI_NIDENT)
        ),
    )

    def __init__(self, filename):
        self.fp = create_memorymapped_fileview(filename)
        sha = calculate_crypto_hash(self.fp.tobytes())

        self.db = model.Model()
        self.session = self.db.session
        self.symbols = SymbolAPI(self)
        self.sections = SectionAPI(self)
        self._images = dict()
        self._sections_by_name = OrderedDict()
        self.asciiCString = CString(encoding = "ascii")
        self._basic_header = ElfParser.BasicHeader.parse(self.fp)
        self.b64 = (self.ei_class == 2)
        if self.ei_data == 1:   # Little-Endian
            offset = 0
        elif self.ei_data == 2: # Big-Endian
            offset = 1
        else:
            raise ValueError("EI_DATA has an invalid value. Got: {}".format(self.ei_data))
        self._endianess = "<" if self.ei_data == 1 else ">"
        datatypes = ElfParser.DATATYPES64.items() if self.b64 else ElfParser.DATATYPES32.items()
        for key, value in datatypes:
            setattr(self, key, value[offset])
        self._parser_extended_header()
        self._parse_section_headers()
        self._parse_program_headers()
        self.create_section_to_segment_mapping()
        for section in self._symbol_sections:
            self._parse_symbol_section(section)
        self.session.commit()

    def _parser_extended_header(self):
        ExtendedHeader = Struct(
            "e_type" / self.Half,        # Object file type
            "e_machine" / self.Half,     # Machine type
            "e_version" / self.Word,     # Object file version
            "e_entry" / self.Addr,       # Entry point address
            "e_phoff" / self.Off,        # Program header offset
            "e_shoff" / self.Off,        # Section header offset
            "e_flags" / self.Word,       # Processor-specific flags
            "e_ehsize" / self.Half,      # ELF header size
            "e_phentsize" / self.Half,   # Size of program header entry
            "e_phnum" / self.Half,       # Number of program header entries
            "e_shentsize" / self.Half,   # Size of section header entry
            "e_shnum" / self.Half,       # Number of section header entries
            "e_shstrndx" / self.Half,    # Section name string table index
        )
        self._extended_header = ExtendedHeader.parse(self.fp[self.EI_NIDENT : ])

    def _parse_section_headers(self):
        SectionHeaders = Struct(
            "sections" / Array(lambda ctx: self.e_shnum,
                "section" / Struct(
                    "sh_name" / self.Word,
                    "sh_type" / self.Word,
                    "sh_flags" / self.Xword,
                    "sh_addr" / self.Addr,
                    "sh_offset" / self.Off,
                    "sh_size" / self.Xword,
                    "sh_link" / self.Word,
                    "sh_info" / self.Word,
                    "sh_addralign" / self.Xword,
                    "sh_entsize" /self.Xword,
                    "allocate" / Computed(lambda ctx: (ctx.sh_type not in (0, 8) and ctx.sh_size > 0)),
                )
            )
        )
        sections = []
        self._symbol_sections = []
        if hasattr(self, 'e_shnum'):
            self._section_headers = SectionHeaders.parse(self.fp[self.e_shoff : ])
            for idx, section in enumerate(self._section_headers.sections):
                if section.allocate:
                    image = self.fp[section.sh_offset : section.sh_offset + section.sh_size]
                else:
                    image = None
                if image is not None:
                    image = image.tobytes()
                self._images[idx] = image
                section.image = image
            for idx, section in enumerate(self._section_headers.sections):
                name = self.get_string(self.e_shstrndx, section.sh_name)
                section.name = name
                self._sections_by_name[name] = section
                image = self._images[idx]
                if section.sh_type == defs.SectionType.SHT_NOTE:
                    note_obj = self._parse_note(image)
                    if note_obj:
                        note = model.Elf_Note(
                            section_name = name, type = note_obj.type, name = note_obj.name, desc = note_obj.desc
                        )
                        sections.append(note)
                elif section.sh_type in (defs.SectionType.SHT_SYMTAB, defs.SectionType.SHT_DYNSYM):
                    self._symbol_sections.append(section)
                elif name == ".comment":
                    cmt_text = self._parse_comment(image)
                    if cmt_text:
                        comment = model.Elf_Comment(text = cmt_text)
                        sections.append(comment)
                db_sec = model.Elf_Section(index = idx, section_name = name, sh_name = section.sh_name,
                    sh_type = section.sh_type, sh_flags = section.sh_flags, sh_addr = section.sh_addr,
                    sh_offset = section.sh_offset, sh_size = section.sh_size, sh_link = section.sh_link,
                    sh_info = section.sh_info, sh_addralign = section.sh_addralign, sh_entsize = section.sh_entsize,
                    section_image = image
                )
                sections.append(db_sec)
            self.session.bulk_save_objects(sections)

    def get_string(self, table_index, entry):
        if entry > len(self._images[table_index]):
            return ""
        name = self.asciiCString.parse(self._images[table_index][entry : ])
        return name

    def _parse_program_headers(self):
        ProgramHeaders = Struct(
            "segments" / Array(lambda ctx: self.e_phnum,
                "segment" / IfThenElse(lambda ctx: self.b64,
                    Struct(
                        "p_type" / self.Word,
                        "p_flags" / self.Word,
                        "p_offset" / self.Off,
                        "p_vaddr" / self.Addr,
                        "p_paddr" / self.Addr,
                        "p_filesz" / self.Xword,
                        "p_memsz" / self.Xword,
                        "p_align" / self.Xword,
                    ),
                    Struct(
                        "p_type" / self.Word,
                        "p_offset" / self.Off,
                        "p_vaddr" / self.Addr,
                        "p_paddr" / self.Addr,
                        "p_filesz" / self.Word,
                        "p_memsz" / self.Word,
                        "p_flags" / self.Word,
                        "p_align" / self.Word,
                    )
                ),
            )
        )
        if hasattr(self, 'e_shnum'):
            #if self.e_shnum:
            #    print("PG_size: {}".format(ProgramHeaders.sizeof() / self.e_phnum))
            self._program_headers = ProgramHeaders.parse(self.fp[self.e_phoff : ])

    def _parse_symbol_section(self, section):
        sh_link = section.sh_link
        symbols = []
        Symbol = Struct(
            "st_name" / self.Word,
            "st_value" / self.Addr,
            "st_size" / self.Word,
            "st_info" / BitStruct(
                "st_bind" / BitsInteger(4),
                "st_type" / BitsInteger(4),
            ),
            "st_other" / Int8ul,
            "symbol_name" / Computed(lambda ctx: self.get_string(sh_link, ctx.st_name)),
            "st_shndx" / self.Half,
        )
        symbol_cache = {}
        num_symbols = len(section.image) // Symbol.sizeof()
        for offset in range(0, len(section.image), Symbol.sizeof()):
            sym = Symbol.parse(section.image[offset : offset + Symbol.sizeof()])
            section_header = None
            if sym.st_shndx in defs.SpecialSections:
                section_name = defs.special_section_name(sym.st_shndx)
            else:
                if not sym.st_shndx in symbol_cache:
                    section_header = self.session.query(model.Elf_Section).\
                        filter(model.Elf_Section.index == sym.st_shndx).first()
                    if section_header:
                        section_name = section_header.section_name
                    else:
                        section_name = str(sym.st_shndx)
            db_sym = model.Elf_Symbol(st_name = sym.st_name, st_value = sym.st_value, st_size = sym.st_size,
                st_bind = sym.st_info.st_bind, st_type = sym.st_info.st_type, st_other = sym.st_other,
                st_shndx = sym.st_shndx, symbol_name = sym.symbol_name, section_name = section_name,
                access = section_header.sh_flags if section_header else 0
            )
            symbols.append(db_sym)
        self.session.bulk_save_objects(symbols)
        self.session.commit()

    def _parse_comment(self, data):
        Line = Struct(
            "line" / CString("ascii"),
            "pos" / Tell
        )
        if not data:
            return ""
        length = len(data)
        result = []
        i = 0
        if data.find(b"\x00") == -1:
            return str(data, "ascii")
        while i < length:
            #print("*** LINE", data[i : ])
            line = Line.parse(data[i : ])
            if line.line:
                result.append(line.line)
            i += line.pos
        return '\n'.join(result)

    def _parse_note(self, data):
        Note = Struct(
            "namesz" / self.Word,
            "descsz" / self.Word,
            "type" / self.Word,
            "name" / Bytes(this.namesz),
            "desc" / Bytes(this.descsz)
        )
        if not data:
            return None
        result = Note.parse(data)
        result.desc = binascii.b2a_hex(result.desc).decode()
        result.name = self.asciiCString.parse(result.name)
        return result

    def debug_sections(self):
        ds = OrderedDict()
        for idx, section in enumerate(self.sections.fetch()):
            name = section.section_name
            if name.startswith('.debug'):
                if name == '.debug_abbrev':
                    pass
                ds[name] = section
        result = OrderedDict()
        for name, section in ds.items():
            result[name]= DebugInfo(section, section.section_image)
        return result

    def section_in_segment1(self, section_header, segment, check_vma, strict):
        has_offset = False
        valid_segment = False
        has_VMA = False
        has_dynamic_size = False
        valid_segment = ((section_header.sh_flags & defs.SectionFlags.SHF_TLS) != 0) and (segment.p_type == defs.PT_TLS or \
            segment.p_type == defs.PT_GNU_RELRO or segment.p_type == defs.PT_LOAD) or \
            ((section_header.sh_flags & defs.SectionFlags.SHF_TLS) == 0 and segment.p_type != defs.PT_TLS and \
            segment.p_type != defs.PT_PHDR)
        has_offset = section_header.sh_type == defs.SectionType.SHT_NOBITS or (section_header.sh_offset >= segment.p_offset \
            and (not strict or (section_header.sh_offset - segment.p_offset <= segment.p_filesz - 1)) \
            and ((section_header.sh_offset - segment.p_offset + self.section_size(section_header, segment)) <= \
            (segment.p_filesz)))
        has_VMA = (not check_vma or (section_header.sh_flags & defs.SectionFlags.SHF_ALLOC) == 0 or (section_header.sh_addr >= \
            segment.p_vaddr and (not strict or (section_header.sh_addr - segment.p_vaddr <= segment.p_memsz - 1)) \
            and ((section_header.sh_addr - segment.p_vaddr + self.section_size(section_header, segment)) <= segment.p_memsz))
        )
        has_dynamic_size = (segment.p_type != defs.PT_DYNAMIC or section_header.sh_size != 0 or segment.p_memsz == 0 \
            or ((section_header.sh_type == defs.SectionType.SHT_NOBITS or (section_header.sh_offset > segment.p_offset \
            and (section_header.sh_offset - segment.p_offset < segment.p_filesz))) \
            and ((section_header.sh_flags & defs.SectionFlags.SHF_ALLOC) == 0 \
            or (section_header.sh_addr > segment.p_vaddr \
            and (section_header.sh_addr - segment.p_vaddr < segment.p_memsz)))) \
        )
        return (valid_segment and has_offset and has_VMA and has_dynamic_size)

    def section_in_segment(self, section_header, segment):
        return self.section_in_segment1(section_header, segment, 1, 0)

    def section_in_segment_strict(self, section_header, segment):
        return self.section_in_segment1(section_header, segment, 1, 1)

    def create_section_to_segment_mapping(self):
        mapping = OrderedDict()
        for idx in range(self.e_phnum):
            segment = self.segments[idx]
            mapping[idx] = []
##
##            for j in range(self.e_shnum):
##                section = self.sections[j]
##                if not self.tbss_special(section, segment) and self.section_in_segment_strict(section, segment):
##                    mapping[idx].append(j)
##
        self.sections_to_segments = mapping
        return self.sections_to_segments

    def tbss_special(self, section_header, segment):
       return ((section_header.sh_flags & defs.SectionFlags.SHF_TLS) != 0 and
           section_header.sh_type == defs.SectionType.SHT_NOBITS and segment.p_type != defs.PT_TLS
        )

    def section_size(self, section_header, segment):
        return 0 if self.tbss_special(section_header, segment) else section_header.sh_size

    def get_basic_header_field(self, name):
        return getattr(self._basic_header.header.fields, name)

    def get_extended_header_field(self, name):
        return getattr(self._extended_header, name)

    @property
    def ei_class(self):
        return self.get_basic_header_field('ei_class')

    @property
    def ei_data(self):
        return self.get_basic_header_field('ei_data')

    @property
    def ei_version(self):
        return self.get_basic_header_field('ei_version')

    @property
    def ei_osabi(self):
        return self.get_basic_header_field('ei_osabi')

    @property
    def ei_abiversion(self):
        return self.get_basic_header_field('ei_abiversion')

    @property
    def header_bytes(self):
        return self._basic_header.header.bytes

    @property
    def e_type(self):
        return self.get_extended_header_field('e_type')

    @property
    def e_machine(self):
        return self.get_extended_header_field('e_machine')

    @property
    def e_version(self):
        return self.get_extended_header_field('e_version')

    @property
    def e_entry(self):
        return self.get_extended_header_field('e_entry')

    @property
    def e_phoff(self):
        return self.get_extended_header_field('e_phoff')

    @property
    def e_shoff(self):
        return self.get_extended_header_field('e_shoff')

    @property
    def e_flags(self):
        return self.get_extended_header_field('e_flags')

    @property
    def e_ehsize(self):
        return self.get_extended_header_field('e_ehsize')

    @property
    def e_phentsize(self):
        return self.get_extended_header_field('e_phentsize')

    @property
    def e_phnum(self):
        return self.get_extended_header_field('e_phnum')

    @property
    def e_shentsize(self):
        return self.get_extended_header_field('e_shentsize')

    @property
    def e_shnum(self):
        return self.get_extended_header_field('e_shnum')

    @property
    def e_shstrndx(self):
        return self.get_extended_header_field('e_shstrndx')

    @property
    def endianess(self):
        return self._endianess

    @property
    def segments(self):
        return self._program_headers['segments']

    @property
    def arm_attributes(self):
        res = self.query(model.Elf_Section).filter(model.Elf_Section.section_name == ".ARM.attributes").first()
        if res:
            return attributes.parse(res.section_image, byteorder = self.endianess)
        else:
            return {}

    @property
    def comment(self):
        comment = self.query(model.Elf_Comment).first()
        if comment:
            return comment.text
        else:
            return None

    @property
    def notes(self):
        notes = self.query(model.Elf_Note).order_by(model.Elf_Note.section_name).all()
        if notes:
            return notes
        else:
            return []

    @property
    def query(self):
        return self.session.query

    def create_image(self, join: bool = True, include_pattern: str = None, exclude_pattern: str = None, callback: callable = None):
        """

        Parameters
        ----------
        join: bool
            Try to join/merge sections.

        include_pattern: str
            Include only sections matching a Python RegEx

        exclude_pattern: str
            Exclude sections matching a Python RegEx

        callback: callable
            Could be used to generate output for your command-line tools.

            Called with:
                - state: "start" | "stop" | "section"
                - section -- current section (only applicable to state "section").

        Returns
        -------
        `objutils.Image`

        Note
        ----
        `include_pattern` and `exclude_pattern` should be used mutually exclusive, unless you know what you are doing.

        Note
        ----
        Look at `scripts/oj_elf_extract.py` to see `create_image()` in action.
        """
        query = sections = self.query(model.Elf_Section)
        query = query.filter(
            model.Elf_Section.flag_alloc == True,
            model.Elf_Section.has_content == True,
        )

        if include_pattern:
            query = query.filter(func.regexp(model.Elf_Section.section_name, include_pattern))

        if exclude_pattern:
            query = query.filter(not_(func.regexp(model.Elf_Section.section_name, exclude_pattern)))

        query = query.order_by(model.Elf_Section.sh_addr)
        result = []
        if callback:
            callback("start", None)
        for section in query.all():
            if callback:
                callback("section", section)
            result.append(Section(section.sh_addr, section.section_image))
        img = Image(result, join = join)
        if callback:
            callback("stop", None)
        return img



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from objutils import loads, dumps, probes
from objutils.hexfile import MetaRecord
from objutils.image import Image
import unittest

import pytest


SIG = b""":B00010A5576F77212044696420796F75207265617B
:B01010E56C6C7920676F207468726F756768206136
:B02010256C6C20746861742074726F75626C652068
:B0300D5F746F207265616420746869733FD1
:B03D00"""

IHEX = b"""
"""

SREC1 = b"""S113B000576F77212044696420796F7520726561D8
S113B0106C6C7920676F207468726F756768206143
S113B0206C20746861742074726F75626C6520742E
S10FB0306F207265616420746869733FCE
S9030000FC"""


SREC2 = b"""S21400B000576F77212044696420796F7520726561D7
S21400B0106C6C7920676F207468726F756768206142
S21400B0206C20746861742074726F75626C6520742D
S21000B0306F207265616420746869733FCD
S804000000FB"""

SREC3 = b"""S3150000B000576F77212044696420796F7520726561D6
S3150000B0106C6C7920676F207468726F756768206141
S3150000B0206C20746861742074726F75626C6520742C
S3110000B0306F207265616420746869733FCC
S70500000000FA"""

SREC4 = b"""S00D000073616D706C652E73313965
S315017FC0003B1879210016C07816C04B10EF87C76C95
S315017FC01080B74618586CEA2102186280EC808C0042
"""

def test_meta_data1():
    img = loads("srec", SREC1)
    assert img.meta == {8: [MetaRecord(format_type=8, address=0, chunk=None)]}

def test_meta_data2():
    img = loads("srec", SREC4)
    assert img.meta == {1: [MetaRecord(format_type=1, address=0, chunk=bytearray(b'sample.s19'))]}

class TestRoundtrip(unittest.TestCase):

    def testLoadsWorks(self):
        image = Image()
        image.insert_section("Wow! Did you really go through al that trouble to read this?", 0xb000)
        image.join_sections()
        self.assertEqual(dumps("srec", image, record_type = 1, s5record = False, start_address = 0x0000), SREC1)


class Test19Probe(unittest.TestCase):

    @pytest.mark.skip
    def testS19ProbeS1(self):
        self.assertEqual(probes(SREC1), "srec")

    @pytest.mark.skip
    def testS19ProbeS2(self):
        self.assertEqual(probes(SREC2), "srec")

    @pytest.mark.skip
    def testS19ProbeS3(self):
        self.assertEqual(probes(SREC3), "srec")


class TestS19Options(unittest.TestCase):

    def createImage(self, record_type, s5record, start_address = None):
        image = Image()
        image.insert_section(range(10), 0x1000)
        image.join_sections()
        return dumps("srec", image, record_type = record_type, s5record = s5record, start_address = start_address)

    def testS19includeS5RecordS1(self):
        self.assertEqual(self.createImage(1, True), b"S10D100000010203040506070809B5\nS5030001FB")

    def testS19includeS5RecordS2(self):
        self.assertEqual(self.createImage(2, True), b"S20E00100000010203040506070809B4\nS504000001FA")

    def testS19includeS5RecordS3(self):
        self.assertEqual(self.createImage(3, True), b"S30F0000100000010203040506070809B3\nS50500000001F9")

    def testS19excludeS5RecordS1(self):
        self.assertEqual(self.createImage(1, False), b"S10D100000010203040506070809B5")

    def testS19excludeS5RecordS2(self):
        self.assertEqual(self.createImage(2, False), b"S20E00100000010203040506070809B4")

    def testS19excludeS5RecordS3(self):
        self.assertEqual(self.createImage(3, False), b"S30F0000100000010203040506070809B3")

    def testS19includeStartAddressS1(self):
        self.assertEqual(self.createImage(1, False, 0x1000), b"S10D100000010203040506070809B5\nS9031000EC")

    def testS19includeStartAddressS2(self):
        self.assertEqual(self.createImage(2, False, 0x1000), b"S20E00100000010203040506070809B4\nS804001000EB")

    def testS19includeStartAddressS3(self):
        self.assertEqual(self.createImage(3, False, 0x1000), b"S30F0000100000010203040506070809B3\nS70500001000EA")


if __name__=='__main__':
    unittest.main()




import pytest

from objutils.checksums import nibble_sum, lrc, rotatedXOR, xor, check
from objutils.checksums import COMPLEMENT_NONE, COMPLEMENT_ONES, COMPLEMENT_TWOS, ROTATE_LEFT, ROTATE_RIGHT


def testNibbleSumCase1():
    assert nibble_sum(range(10)) == 45

def testNibbleSumCase2():
    assert nibble_sum(range(100)) == 222

    # check that s.split fails when the separator is not a string
    #with self.assertRaises(TypeError):
    #    s.split(2)

def testXorInverted():
    assert xor([0xbc, 0xaf, 0xfe, 0x11, 0x52, 0x60, 0x80], True) == 0xb1

def testXorNotInverted():
    assert xor([0xbc, 0xaf, 0xfe, 0x11, 0x52, 0x60, 0x80], False) == 0x4e



def testLRCCase1():
    assert lrc(range(10), 8, COMPLEMENT_NONE) == 45

def testLRCCase2():
    assert lrc(range(100), 8, COMPLEMENT_NONE) == 86

def testLRCCase3():
    assert lrc(range(10), 8, COMPLEMENT_ONES) == 210

def testLRCCase4():
    assert lrc(range(100), 8, COMPLEMENT_ONES) == 169

def testLRCCase5():
    assert lrc(range(10), 8, COMPLEMENT_TWOS) == 211

def testLRCCase6():
    assert lrc(range(100), 8, COMPLEMENT_TWOS) == 170

def testLRCCase7():
    assert lrc(range(10), 16, COMPLEMENT_NONE) == 45

def testLRCCase8():
    assert lrc(range(100), 16, COMPLEMENT_NONE) == 4950

def testLRCCase9():
    assert lrc(range(10), 16, COMPLEMENT_ONES) == 65490

def testLRCCase10():
    assert lrc(range(100), 16, COMPLEMENT_ONES) == 60585

def testLRCCase11():
    assert lrc(range(10), 16, COMPLEMENT_TWOS) == 65491

def testLRCCase12():
    assert lrc(range(100), 16, COMPLEMENT_TWOS) == 60586

def testLRCCase13():
    assert lrc(range(10), 32, COMPLEMENT_NONE) == 45

def testLRCCase14():
    assert lrc(range(100), 32, COMPLEMENT_NONE) == 4950

def testLRCCase15():
    assert lrc(range(10), 32, COMPLEMENT_ONES) == 4294967250

def testLRCCase16():
    assert lrc(range(100), 32, COMPLEMENT_ONES) == 4294962345

def testLRCCase17():
    assert lrc(range(10), 32, COMPLEMENT_TWOS) == 4294967251

def testLRCCase18():
    assert lrc(range(100), 32, COMPLEMENT_TWOS) == 4294962346



def testRXORCase1():
    assert rotatedXOR(range(10), 8, ROTATE_LEFT) == 74

def testRXORCase2():
    assert rotatedXOR(range(100), 8, ROTATE_LEFT) == 66

def testRXORCase3():
    assert rotatedXOR(range(10), 8, ROTATE_RIGHT) == 186

def testRXORCase4():
    assert rotatedXOR(range(100), 8, ROTATE_RIGHT) == 3

def testRXORCase5():
    assert rotatedXOR(range(10), 16, ROTATE_LEFT) == 74

def testRXORCase6():
    assert rotatedXOR(range(100), 16, ROTATE_LEFT) == 66

def testRXORCase7():
    assert rotatedXOR(range(10), 16, ROTATE_RIGHT) == 186

def testRXORCase8():
    assert rotatedXOR(range(100), 16, ROTATE_RIGHT) == 3

def testRXORCase9():
    assert rotatedXOR(range(10), 32, ROTATE_LEFT) == 74

def testRXORCase10():
    assert rotatedXOR(range(100), 32, ROTATE_LEFT) == 66

def testRXORCase11():
    assert rotatedXOR(range(10), 32, ROTATE_RIGHT) == 186

def testRXORCase12():
    assert rotatedXOR(range(100), 32, ROTATE_RIGHT) == 3

TEST = bytes((
    0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08,
    0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F, 0x10,
    0xF1, 0xF2, 0xF3, 0xF4, 0xF5, 0xF6, 0xF7, 0xF8,
    0xF9, 0xFA, 0xFB, 0xFC, 0xFD, 0xFE, 0xFF, 0x00,
))


def testAdd11():
    assert check(TEST, "CHK_ADD_11") == 0x10


def testAdd12():
    assert check(TEST, "CHK_ADD_12") == 0x0F10


def testAdd14():
    assert check(TEST, "CHK_ADD_14") == 0x00000F10


def testAdd22():
    assert check(TEST, "CHK_ADD_22") == 0x1800


def testAdd24():
    assert check(TEST, "CHK_ADD_24") == 0x00071800


def testAdd44():
    assert check(TEST, "CHK_ADD_44") == 0x140C03F8


def testCrc16():
    assert check(TEST, "CHK_CRC_16") == 0xC76A


def testCrc16Ccitt():
    assert check(TEST, "CHK_CRC_16_CCITT") == 0x9D50


@pytest.mark.parametrize("x, expected",
    [
        ([0x00, 0x00, 0x00, 0x00], 0x84c0),
        ([0xf2, 0x01, 0x83], 0xD374),
        ([0x0f, 0xaa, 0x00, 0x55], 0x2023),
        ([0x00, 0xff, 0x55, 0x11], 0xB8F9),
        ([0x33, 0x22, 0x55, 0xaa, 0xbb, 0xcc, 0xdd, 0xee, 0xff], 0xF53F),
        ([0x92, 0x6b, 0x55], 0x0745),
        ([0xff, 0xff, 0xff, 0xff], 0x1D0F),
    ]
)
def test_crc16_ccitt_ar(x, expected):
    assert check(x, "CHK_CRC_16_CCITT") == expected

@pytest.mark.parametrize("x, expected",
    [
        ((0x00, 0x00, 0x00, 0x00), 0x2144DF1C),
        ([0xf2, 0x01, 0x83], 0x24AB9D77),
        ([0x0f, 0xaa, 0x00, 0x55], 0xB6C9B287),
        ([0x00, 0xff, 0x55, 0x11], 0x32A06212),
        ([0x33, 0x22, 0x55, 0xaa, 0xbb, 0xcc, 0xdd, 0xee, 0xff], 0xB0AE863D),
        ([0x92, 0x6b, 0x55], 0x9CDEA29B),
        ([0xff, 0xff, 0xff, 0xff], 0xFFFFFFFF),
    ]
)
def test_crc32_ar(x, expected):
    assert check(x, "CHK_CRC_32") == expected

def testCrc32():
    assert check(TEST, "CHK_CRC_32") == 0x89CD97CE


def testUserDefined():
    with pytest.raises(NotImplementedError):
        check(TEST, "CHK_USER_DEFINED")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""DWARF4 Parser.
"""

__version__ = "0.1.0"
__copyright__ = """
    objutils - Object file library for Python.

   (C) 2010-2021 by Christoph Schueler <github.com/Christoph2,
                                        cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

from collections import namedtuple, OrderedDict
import io

from construct import CString, Enum, If, Struct
from construct import this
from construct import Computed, Pass, Switch, Tell
from construct import Int16ul, Int32ul, Int64ul, Int8ul
from construct import Int16ub, Int32ub, Int64ub

from objutils.dwarf import constants
from objutils.dwarf.encoding import ULEB, SLEB, Address, StrP, Block1, BlockUleb, One

Abbreviation = namedtuple('Abbreviation', 'tag children attrs')

ident = lambda x: x

class DwarfProcessor:
    """
    """
    DATATYPES32 = {
        "Addr":   (Int32ul, Int32ub),   # 4 - Unsigned program address
    }

    UTF8String = CString(encoding = "utf8")

    def __init__(self, sections, b64, endianess):
        self.b64 = b64
        self.endianess = 1 if endianess == "<" else 0
        debug_sections = sections.fetch(name_pattern = ".debug*")
        self.debug_sections = {section.section_name: section for section in debug_sections}
        if ".debug_str" in self.debug_sections:
            self.strings = self.debug_sections[".debug_str"].section_image
        else:
            self.strings = b""

        #print(self.strings.read())
        self.installReaders()

    def installReaders(self):
        BASIC_READERS = {
            "u8":   (Int8ul, Int8ul),
            "u16":  (Int16ul, Int16ub),
            "u32":  (Int32ul, Int32ub),
            "u64":  (Int64ul, Int64ub)
        }
        idx = 0 if self.endianess == 1 else 1
        for name, reader in BASIC_READERS.items():
            setattr(self, name, reader[idx])

    def get_string(self, offset):
        #self.strings.seek(offset)
        #result = self.UTF8String.parse_stream(self.strings)
        result = str(self.debug_sections[".debug_str"])[offset : offset + 25]
        return bytes(result, encoding = "ascii")


    def do_abbrevs(self):
        section = self.debug_sections['.debug_abbrev']
        image = section.section_image
        length = len(section.section_image)
        Abbrevation = Struct(
            "start" / Tell,
            "code" / ULEB,
            "details" / If(lambda ctx: ctx.code != 0,
            Struct(
                    "tag" / ULEB,
                    "children" / self.u8,
                )
            ),
            "stop" / Tell,
        )
        Attribute = Struct(
            "start" / Tell,
            "attrValue" / ULEB,
            "formValue" / ULEB,
            "stop" / Tell,
            "next"/ Computed((this.attrValue != 0) and (this.formValue != 0)),
        )
        offset = 0
        result = OrderedDict()
        kOffs = offset
        while True:
            print("Offset: {}".format(offset))
            abbrev = Abbrevation.parse(image[offset : ])
            if abbrev.code != 0:
                print("{} {} {}".format(abbrev.code, constants.Tag(abbrev.details.tag), abbrev.details.children == 1))
            key = (kOffs, abbrev.code, )
            offset += abbrev.stop - abbrev.start
            if offset >= length:
                break
            if abbrev.code == 0:
                kOffs = offset
                result[key] = Abbreviation(0, False, [])
                continue
            result[key] = Abbreviation(constants.Tag(abbrev.details.tag), (abbrev.details.children == 1), [])
            while True:
                attr = Attribute.parse(image[offset : ])

                if attr.attrValue != 0 and attr.formValue != 0:
                    result[key].attrs.append((constants.AttributeEncoding(attr.attrValue), constants.AttributeForm(attr.formValue)))
                    print("   {} {}".format(constants.AttributeEncoding(attr.attrValue), constants.AttributeForm(attr.formValue)))
                offset += (attr.stop - attr.start)
                if not attr.next:
                    break
        self.abbreviations = result

    def do_mac_info(self):
        if not '.debug_macinfo' in self.debug_sections:
            return
        section = self.debug_sections['.debug_macinfo']
        image = io.BytesIO(section.section_image)
        length = len(section.section_image)

        MacInfo = Struct(
            "start" / Tell,
            "macType" / Enum(ULEB, default = ident,
                DW_MACINFO_define  = 0x01,
                DW_MACINFO_undef  = 0x02,
                DW_MACINFO_start_file  = 0x03,
                DW_MACINFO_end_file  = 0x04,
                DW_MACINFO_vendor_ext  = 0xff,
            ),
            "parameters" / Switch(this.macType, {
                "DW_MACINFO_define": Struct("lineNumber" / ULEB, "value" / CString(encoding = "ascii")),
                "DW_MACINFO_undef": Struct("lineNumber" / ULEB, "value" / CString(encoding = "ascii")),
                "DW_MACINFO_start_file": Struct("lineNumber" / ULEB, "fileNumber" / ULEB),
                "DW_MACINFO_end_file": Pass,
                "DW_MACINFO_vendor_ext": Struct("constant" / ULEB, "value" / CString(encoding = "ascii")),
            },
                default = Pass,
            ),
            "stop" / Tell,
        )
        offset = 0
        while True:
            macInfo = MacInfo.parse_stream(image)
            offset += macInfo.stop - macInfo.start
            print(offset, macInfo)
            if offset >= length - 1:
                break

    def get_form_readers(self, addressSize):
        return {
            constants.AttributeForm.DW_FORM_string:       CString(encoding = "utf8"),
            constants.AttributeForm.DW_FORM_udata:        ULEB,
            constants.AttributeForm.DW_FORM_sdata:        SLEB,
            constants.AttributeForm.DW_FORM_data1:        self.u8,
            constants.AttributeForm.DW_FORM_data2:        self.u16,
            constants.AttributeForm.DW_FORM_data4:        self.u32,
            constants.AttributeForm.DW_FORM_data8:        self.u64,
            constants.AttributeForm.DW_FORM_addr:         Address(addressSize, self.endianess),
            constants.AttributeForm.DW_FORM_block:        BlockUleb,
            constants.AttributeForm.DW_FORM_block1:       Block1,
            constants.AttributeForm.DW_FORM_block2:       'block2',
            constants.AttributeForm.DW_FORM_block4:       'block4',
            constants.AttributeForm.DW_FORM_flag:         self.u8,
            constants.AttributeForm.DW_FORM_ref_addr:     self.u32,
            constants.AttributeForm.DW_FORM_ref1:         self.u8,
            constants.AttributeForm.DW_FORM_ref2:         self.u16,
            constants.AttributeForm.DW_FORM_ref4:         self.u32,
            constants.AttributeForm.DW_FORM_ref8:         self.u64,
            constants.AttributeForm.DW_FORM_ref_udata:    ULEB,
            constants.AttributeForm.DW_FORM_strp:         StrP(self.strings, self.endianess),
            constants.AttributeForm.DW_FORM_indirect:     None, # TODO: uleb value, that represents its form!
            constants.AttributeForm.DW_FORM_sec_offset:   self.u32,
            constants.AttributeForm.DW_FORM_exprloc:      BlockUleb,
            constants.AttributeForm.DW_FORM_flag_present: One,
            constants.AttributeForm.DW_FORM_ref_sig8:     self.u64,
        }

    def process_attributes(self, image, readers, size, abbrevOffset):
        Attribute = Struct(
            "start" / Tell,
            "attr" / ULEB,
            "stop" / Tell,
            "size" / Computed(this.stop - this.start),
        )
        level = 0
        offset = 0
        lastAttr = False
        while True:
            start = image.tell()
            attr = Attribute.parse_stream(image)
            offset += attr.size
            abbr = self.abbreviations.get((abbrevOffset, attr.attr))
            if attr.attr == 0 or not abbr.tag:
                print("<{}><{:02x}>: {}".format(level, start, "Abbrev Number: 0"))
                level -= 1
                if lastAttr:
                    break
            else:
                #print(dir(abbr))
                print(abbr.tag.MAP)
                #print("<{}><{:02x}>: Abbrev Number: {} ({})".format(level, start, attr.attr, abbr.tag.name))
                print("<{}><{:02x}>: Abbrev Number: {} ()".format(level, start, attr.attr, ))
                for enc, form in abbr.attrs:
                    reader = readers.get(form)
                    start = image.tell()
                    if form != constants.DW_FORM_flag_present:
                        value = reader.parse_stream(image)
                    else:
                        value = 1
                    startValue = "<{:x}>".format(start)
                    print("   {:7} {:20}: {}".format(startValue, enc.name, value))
                    stop = image.tell()
                    offset += (stop - start)
                    if offset >= size - 1:
                        lastAttr = True
                pos = image.tell()
            if hasattr(abbr, "children") and abbr.children:
                level += 1

    def process_compile_unit(self, image):
        CompileUnit = Struct(
            "start" / Tell,
            "unit_length" / self.u32,
            "version" / self.u16,
            "debug_abbrev_offset" / self.u32,
            "address_size" / self.u8,
            "stop" / Tell,
            "size" / Computed(this.stop - this.start),
        )
        startPos = image.tell()
        cu = CompileUnit.parse_stream(image)
        print("   Compilation Unit @ offset 0x{:x}:".format(cu.start))
        print("   Length:        0x{:x} (32-bit)".format(cu.unit_length))
        print("   Version:       2".format(cu.version))
        print("   Abbrev Offset: 0x{:x}".format(cu.debug_abbrev_offset))
        print("   Pointer Size:  {}".format(cu.address_size))
        stopPos = image.tell()
        return cu

    def do_dbg_info(self):
        if not '.debug_info' in self.debug_sections:
            return
        section = self.debug_sections['.debug_info']
        image = io.BytesIO(section.section_image)
        length = len(section.section_image)
        DbgInfo = Struct(
            "start" / Tell,
            "unit_length" / self.u32,
            "version" / self.u16,
            "debug_abbrev_offset" / self.u32,
            "address_size" / self.u8,
            "stop" / Tell,
            "size" / Computed(this.stop - this.start),
        )
        Attribute = Struct(
            "start" / Tell,
            "attr" / ULEB,
            "stop" / Tell,
            "size" / Computed(this.stop - this.start),
        )
        offset = 0
        finished = False
        while True:
            pos = image.tell()
            if pos >= length - 1:
                break
            dbgInfo = DbgInfo.parse_stream(image) # CU
            print("   Compilation Unit @ offset 0x{:x}:".format(dbgInfo.start))
            print("   Start:         {:08x}".format(dbgInfo.start))
            print("   Stop:          {:08x}".format(dbgInfo.stop))
            print("   Size:          {:08x}".format(dbgInfo.size))
            print("   Length:        0x{:x} (32-bit)".format(dbgInfo.unit_length))
            print("   Version:       2".format(dbgInfo.version))
            print("   Abbrev Offset: 0x{:x}".format(dbgInfo.debug_abbrev_offset))
            print("   Pointer Size:  {}".format(dbgInfo.address_size))
            pos = 0
            offset += dbgInfo.stop - dbgInfo.start
            formReaders = self.get_form_readers(dbgInfo.address_size)
            print("Pos, Length", pos, dbgInfo.unit_length)
            if pos >= dbgInfo.unit_length:
                break
            while True:
                start = image.tell()
                attr = Attribute.parse_stream(image)
                abbr = self.abbreviations.get((dbgInfo.debug_abbrev_offset, attr.attr))
                if not abbr:
                    print("<{:02x}>: {}".format(start, "Abbrev Number: 0"))
                else:
                    #print("<{:02x}>: Abbrev Number: {} ({})".format(start, enc.name, value))
                    for enc, form in abbr.attrs:
                        reader = formReaders.get(form)
                        if reader is None:
                            print("*EF", enc, form, start, attr, abbr)
                        start = image.tell()
                        value = reader.parse_stream(image)
                        print("    <{:02x}> {}: {}".format(start, enc, value))
                        stop = image.tell()
                    offset += (attr.stop - attr.start)
                    pos = image.tell()
                    #if pos >= 0x8727:
                    #    print("chk")
                    if pos >= dbgInfo.unit_length:
                        image.seek(image.tell() + 1)
                        break

    def pubnames(self):
        if not '.debug_pubnames' in self.debug_sections:
            return
        section = self.debug_sections['.debug_pubnames']
        image = io.BytesIO(section.section_image)
        length = len(section.section_image)

        Header = Struct(
            "start" / Tell,
            "unit_length" / self.u32,
            "version" / self.u16,
            "debug_info_offset" / self.u32,
            "debug_info_length" / self.u32,
            "stop" / Tell,
        )
        Entry = Struct (
            "start" / Tell,
            "offset" / self.u32,
            "name" / CString(encoding = "ascii"),
            "stop" / Tell,
        )
        offset = 0
        finished = False
        while True:
            if finished:
                break
            header = Header.parse_stream(image)
            offset += header.stop - header.start
            print(offset, header)
            if offset >= length - 1:
                break
            while True:
                entry = Entry.parse_stream(image)
                offset += entry.stop - entry.start
                print("ENTRY", entry)
                if entry.offset == 0:
                    break
                if entry.stop >= header.unit_length:
                    finished = True
                    break

    def aranges(self):
        if not '.debug_aranges' in self.debug_sections:
            return
        section = self.debug_sections['.debug_aranges']
        image = io.BytesIO(section.section_image)
        length = len(section.section_image)
        print("ARANGES")
        Header = Struct(
            "start" / Tell,
            "unit_length" / self.u32,
            "version" / self.u16,
            "debug_info_offset" / self.u32,
            "address_size" / self.u8,
            "segment_size" / self.u8,
            "stop" / Tell,
        )
        Entry = Struct (
            "start" / Tell,
            "length" / self.u32,
            "address" / self.u32,
            "stop" / Tell,
        )
        offset = 0
        finished = False
        while True:
            if finished:
                break
            header = Header.parse_stream(image)
            offset += header.stop - header.start
            print(offset, header)
            if offset >= length - 1:
                break
            while True:
                entry = Entry.parse_stream(image)
                offset += entry.stop - entry.start
                print("ENTRY", entry)
                if entry.address == 0 and entry.length == 0:
                    break
                if entry.stop >= header.unit_length:
                    finished = True
                    break

"""
    dbSecs = ep.debugSections()

    dp = DwarfProcessor(dbSecs, ep.b64, ep.endianess)
    dp.pubnames()
    dp.aranges()
    dp.doAbbrevs()
    dp.processDebugInfo()
    #dp.doMacInfo()
    values = (0x5b6e, 0xcd6, 0x5b6e)
"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__version__ = "0.1.0"

__copyright__ = """
    objutils - Object file library for Python.

   (C) 2010-2019 by Christoph Schueler <cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import objutils.hexfile as hexfile

DATA0    = 1
DATA1    = 2
DATA2    = 3
DATA3    = 4


class Reader(hexfile.Reader):

    FORMAT_SPEC = (
        (DATA0,  "!MAAAA DD"),
        (DATA1,  "\?MAAAA DD"),
        (DATA2,  "AAAA DD"),
        (DATA3,  "DD"),
    )
    previous_address = 0
    previous_length = 0

    def check_line(self, line, format_type):
        return True

    def is_data_line(self, line, format_type):
        if format_type == DATA3:
            if line.junk in ("!M", "?M"):   # Startsymbol, address ommited.
                return False
            line.address = self.previous_address + self.previous_length
            self.previous_address = line.address
            self.previous_length = len(line.chunk)
        else:
            if hasattr(line, 'chunk'):
                length = len(line.chunk)
            else:
                length = 0
            self.previous_address = line.address
            self.previous_length = length
        return format_type in (DATA0, DATA1, DATA2, DATA3)


class Writer(hexfile.Writer):

    MAX_ADDRESS_BITS = 16

    def compose_row(self, address, length, row):
        line = "!M{0:04X} {1}".format(address, Writer.hex_bytes(row))
        return line



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__version__ = "0.1.0"

__copyright__ = """
    objutils - Object file library for Python.

   (C) 2010-2019 by Christoph Schueler <github.com/Christoph2,
                                        cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import re
import objutils.hexfile as hexfile

DATA    = 1
EOF     = 2

NULLS = re.compile(r'\0*\s*!M\s*(.*)', re.DOTALL | re.M)


class Reader(hexfile.Reader):

    #data = re.sub('\0*$', ';\n:0000', NULLS.match(inFile.read()).group(1), 1)  # FIXME!!!
    FORMAT_SPEC = (
        (DATA, "AAAA DD;"),
        (EOF, ":0000")
    )

    def check_line(self, line, format_type):
        if format_type == DATA:
            line.length = len(line.chunk)

    def is_data_line(self, line, format_type):
        return format_type == DATA


class Writer(hexfile.Writer):
    SEPARATOR = "{0!s}\x0d\x0a".format(('\x00' * 48))
    MAX_ADDRESS_BITS = 16

    def compose_row(self, address, length, row):
        return "{0:04X} {1!s};".format(address, Writer.hex_bytes(row))

    def compose_header(self, meta):
        return "{0!s}!M".format(Writer.SEPARATOR)

    def compose_footer(self, meta):
        return "{0!s}".format(Writer.SEPARATOR)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__version__ = "0.1.0"

__copyright__ = """
    objutils - Object file library for Python.

   (C) 2010-2019 by Christoph Schueler <cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

##
##  ASCII Space Hex format.
##

from functools import partial
import re
import objutils.hexfile as hexfile
from objutils.checksums import lrc

STX = '\x02'
ETX = '\x03'

DATA = re.compile(r'(?:.*?\02)(?P<chunks>.*)(?:\03)\s*(?:\$\$(?P<checksum>[0-9a-zA-Z]{2,4})[,.])?', re.DOTALL | re.MULTILINE)
ADDRESS = re.compile(r'\$A(?P<value>[0-9a-zA-Z]{2,8})[,.]\s*')
LINE_SPLIITER = re.compile(r"[ %,']")

checksum = partial(lrc, width = 16)

class Reader(hexfile.ASCIIHexReader):
    """
    """
    VALID_CHARS = re.compile(r"^[a-fA-F0-9 %,\'\$\x02\x03\n\r]*$")

    def __init__(self, address_pattern = r'^(?:(?P<stx>[\x02])\s+)?\$A(?P<address>[0-9a-zA-Z]{2,8})[,.]\s*$',
                 data_pattern = r'^(?:[0-9a-zA-Z]{{2,4}}[{0}]?)*\s*$', etx_pattern = r'^q.*$'):
        super(Reader, self).__init__(address_pattern, data_pattern, etx_pattern, separators = ", %'")


class Writer(hexfile.ASCIIHexWriter):

    MAX_ADDRESS_BITS = 16
    ADDRESS_DESIGNATOR = '$A'

    def __init__(self, address_designator = '$A'):
        super(Writer, self).__init__(address_designator)

    def compose_header(self, meta):
        self.checksum = 0
        self.previous_address = None
        line ="{0} ".format(STX)
        return line

    def compose_footer(self, meta):
        line = "{0}$${1:04X},".format(ETX, self.checksum % 65536)
        return line

    def row_callout(self, address, length, row):
        self.checksum += checksum(row)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__version__ = "0.1.0"

__copyright__ = """
    objutils - Object file library for Python.

   (C) 2010-2019 by Christoph Schueler <cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import re

import objutils.hexfile as hexfile
import objutils.checksums as checksums
import objutils.utils as utils


DATA    = 1
SYMBOL  = 2
EOF     = 3


class Reader(hexfile.Reader):

    VALID_CHARS = re.compile(r"^[a-zA-Z0-9_ %\n\r]*$")    # We need to consider symbol information.

    FORMAT_SPEC = (
        (DATA,      "%LL6CCAAAAADD"),
        (SYMBOL,    "%LL3CCU"),
        (EOF,      "%LL8CCAAAAADD"),
    )

    def check_line(self, line, format_type):
        if format_type == DATA:
            line.length = (line.length / 2) - 5
            checksum = checksums.nibble_sum(utils.make_list(utils.int_to_array(line.address), 6, ((line.length + 5) * 2), line.chunk))
            if line.length != len(line.chunk):
                raise hexfile.InvalidRecordLengthError("Byte count doesn't match length of actual data.")
            if line.checksum!=checksum:
                raise hexfile.InvalidRecordChecksumError()
        elif format_type == SYMBOL:
            checksum = checksums.nibble_sum(utils.make_list(3, ((line.length + 5) * 2), [ord(b) for b in line.chunk]))
            chunk = line.chunk.strip()
            address = int(chunk[-4 : ], 16)
            line.address = address
            #if line.checksum!=checksum:
            #    raise hexfile.InvalidRecordChecksumError()

    def is_data_line(self, line, format_type):
        return format_type == DATA

    def parseData(self, line, format_type):
        return format_type != SYMBOL

class Writer(hexfile.Writer):

    MAX_ADDRESS_BITS = 24

    def compose_row(self, address, length, row):
        checksum = checksums.nibble_sum(utils.make_list(utils.int_to_array(address), 6, ((length + 5) * 2), row))

        line = "%{0:02X}6{1:02X}{2:04X}{3!s}".format((length + 5) * 2, checksum, address, Writer.hex_bytes(row) )
        return line



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__version__ = "0.1.0"

__copyright__ = """
    objutils - Object file library for Python.

   (C) 2010-2020 by Christoph Schueler <cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

class FileCorruptedError(Exception): pass

class PrematureEndOfFileError(Exception): pass

class InvalidAddressError(Exception):
    """Raised if address information is out of range.
    """

class FeatureNotAvailableError(Exception):
    """Raised if something is missing, e.g. Numpy for certain operations.
    """




#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Package entry-point.

Registers CODECS and implements an interface to them.

- :func:`load`
- :func:`loads`
- :func:`probe`
- :func:`probes`
- :func:`dump`
- :func:`dumps`

The first parameter is always the codec name.
"""

__version__ = "0.4.7"

__copyright__ = """
    objutils - Object file library for Python.

   (C) 2010-2020 by Christoph Schueler <cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""


from objutils.registry import registry
import objutils.binfile
import objutils.sig
import objutils.srec
import objutils.titxt
import objutils.emon52
import objutils.etek
import objutils.fpc
import objutils.ihex
import objutils.mostec
import objutils.rca
import objutils.tek
import objutils.cosmac
import objutils.ash
import objutils.shf

from objutils.image import Image
from objutils.section import Section

registry.register('bin', objutils.binfile.Reader, objutils.binfile.Writer, "Plain binary format.")
registry.register('binzip', objutils.binfile.BinZipReader, objutils.binfile.BinZipWriter, "Zipped binary format.")
registry.register('sig', objutils.sig.Reader, objutils.sig.Writer, "Signetics format.")
registry.register('srec', objutils.srec.Reader, objutils.srec.Writer, "Motorola S-Records (a.k.a. S19).")
registry.register('titxt', objutils.titxt.Reader, objutils.titxt.Writer, "Texas Instruments MSP430 text format.")
registry.register('emon52', objutils.emon52.Reader, objutils.emon52.Writer, "Elektor Monitor (EMON52) file format.")
registry.register('etek', objutils.etek.Reader, objutils.etek.Writer, "Extended Tektonix format.")
registry.register('fpc', objutils.fpc.Reader, objutils.fpc.Writer, "Four packed code file format.")
registry.register('ihex', objutils.ihex.Reader, objutils.ihex.Writer, "Intel IHex format.")
registry.register('mostec', objutils.mostec.Reader, objutils.mostec.Writer, "MOSTech format.")
registry.register('rca', objutils.rca.Reader, objutils.rca.Writer, "RCA format.")
registry.register('tek', objutils.tek.Reader, objutils.tek.Writer, "Tektonix format.")
registry.register('cosmac', objutils.cosmac.Reader, objutils.cosmac.Writer, "RCA COSMAC format.")
registry.register('ash', objutils.ash.Reader, objutils.ash.Writer, "ASCII hex space formats.")
registry.register('shf', objutils.shf.Reader, objutils.shf.Writer, "S Hexdump Format (rfc4149).")

def load(codec_name, *args, **kws):
    """Load hex data from file.

    Parameters
    ----------
    codec_name: str
        Name of a registered codec.

    Returns
    -------
    class:`Image`
    """
    return registry.get(codec_name).Reader().load(*args, **kws)

def loads(codec_name, *args, **kws):
    """Load hex data from bytes.

    Parameters
    ----------
    codec_name: str
        Name of a registered codec.

    Returns
    -------
    class:`Image`
    """

    return registry.get(codec_name).Reader().loads(*args, **kws)

def probe(*args, **kws):
    """Try to guess codec from file.

    Parameters
    ----------
    codec_name: str
        Name of a registered codec.

    Returns
    -------
    str
    """

    found = False
    for _, codec in registry._codecs.items():
        reader = codec.Reader()
        found = reader.probe(*args, **kws)
        if found:
            break
    return reader.codec_name if found else None

def probes(*args, **kws):
    """Try to guess codec from bytes.

    Parameters
    ----------
    codec_name: str
        Name of a registered codec.

    Returns
    -------
    str
    """

    found = False
    for _, codec in registry:
        reader = codec.Reader()
        found = reader.probes(*args, **kws)
        if found:
            break
    return reader.codec_name if found else None

def dump(codec_name, *args, **kws):
    """Save hex data to file.

    Parameters
    ----------
    codec_name: str
        Name of a registered codec.

    Returns
    -------
    bytes
    """

    registry.get(codec_name).Writer().dump(*args, **kws)

def dumps(codec_name, *args, **kws):
    """Save hex data to bytes.

    Parameters
    ----------
    codec_name: str
        Name of a registered codec.

    Returns
    -------
    bytes
    """

    return registry.get(codec_name).Writer().dumps(*args, **kws)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__version__ = "0.1.0"

__copyright__ = """
    objutils - Object file library for Python.

   (C) 2010-2020 by Christoph Schueler <github.com/Christoph2,
                                        cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import mmap
import re
import sqlite3

from sqlalchemy import (MetaData, types, orm, event, create_engine,
    Column, and_, not_
)
from sqlalchemy.engine import Engine
from sqlalchemy.ext.declarative import declarative_base, declared_attr
from sqlalchemy.ext.hybrid import hybrid_method, hybrid_property

from objutils.elf import defs

CACHE_SIZE      = 4 # MB
PAGE_SIZE       = mmap.PAGESIZE

Base = declarative_base()

class MixInBase(object):

    @declared_attr
    def __tablename__(cls):
        return cls.__name__.lower()

    def __repr__(self):
        columns = [c.name for c in self.__class__.__table__.c]
        result = []
        for name, value in [(n, getattr(self, n)) for n in columns]:
            if isinstance(value, str):
                result.append("{} = '{}'".format(name, value))
            else:
                result.append("{} = {}".format(name, value))
        return "{}({})".format(self.__class__.__name__, ", ".join(result))


class RidMixIn(MixInBase):
    rid = Column("rid", types.Integer, primary_key = True)


"""
class SLBigInteger(BigInteger):
    pass

@compiles(SLBigInteger, 'sqlite')
def bi_c(element, compiler, **kw):
    return "INTEGER"

@compiles(SLBigInteger)
def bi_c(element, compiler, **kw):
    return compiler.visit_BIGINT(element, **kw)


table = Table(
    "my_table", metadata,
    Column("id", SLBigInteger(), primary_key=True)
)
"""

def StdInteger(default = 0, primary_key = False, unique = False, nullable = False, index = False):
    return Column(types.INTEGER, default = default, nullable = nullable,    # PassiveDefault(str(default))
        primary_key = primary_key, unique = unique, index = index)
    #types.INTEGER

def StdNumeric(default = 0, primary_key = False, unique = False, nullable = False, index = False):
    return Column(types.NUMERIC(20, 0), default = default, nullable = nullable,
        primary_key = primary_key, unique = unique, index = index)

def StdBigInt(default = 0, primary_key = False, unique = False, nullable = False, index = False):
    return Column(types.INTEGER, default = default, nullable = nullable,
        primary_key = primary_key, unique = unique, index = index)

    #  Column("id", types.BigInteger().with_variant(types.BigInteger, "sqlite"), primary_key=True)

def StdFloat(default = 0.0, primary_key = False, unique = False, nullable = False):
    return Column(types.Integer, default = default, nullable = nullable,
        primary_key = primary_key, unique = unique)


class MetaData(Base, RidMixIn):
    """

    """
    sha = Column(types.VARCHAR) # hashlib.sha3_512()


class Elf_Section(Base, RidMixIn):
    """

    """
    index = Column(types.Integer, unique = True)
    sh_name = StdBigInt()
    sh_type = StdBigInt()
    sh_flags = StdBigInt()
    sh_addr = StdBigInt()
    sh_offset = StdBigInt()
    sh_size = StdBigInt()
    sh_link = StdBigInt()
    sh_info = StdBigInt()
    sh_addralign = StdBigInt()
    sh_entsize = StdBigInt()

    section_name = Column(types.VARCHAR, index = True)
    section_image = Column(types.BLOB)

    @hybrid_property
    def is_debug(self):
        return self.section_name.starts_with(".debug")

    @hybrid_property
    def is_dwo(self):
        return self.section_name.starts_with(".dwo")

    @hybrid_property
    def has_content(self):
        return self.sh_type not in (defs.SectionType.SHT_NOBITS, defs.SectionType.SHT_NULL) and (self.sh_size > 0)

    @has_content.expression
    def has_content(self):
        return and_(not_(self.sh_type.in_((defs.SectionType.SHT_NOBITS, defs.SectionType.SHT_NULL))), (self.sh_size > 0))

    @hybrid_method
    def get_flags(self):
        return self.sh_flags

    @hybrid_method
    def test_flags(self, mask):
        return self.get_flags() & mask == mask

    @test_flags.expression
    def test_flags(cls, mask):
        return cls.get_flags().op("&")(mask) == mask

    @hybrid_property
    def section_type(self):
        return defs.SectionType(self.sh_type)

    @hybrid_property
    def flag_writeable(self):
        return self.test_flags(defs.SectionFlags.SHF_WRITE)

    @hybrid_property
    def flag_executable(self):
        return self.test_flags(defs.SectionFlags.SHF_EXECINSTR)

    @hybrid_property
    def flag_alloc(self):
        return self.test_flags(defs.SectionFlags.SHF_ALLOC)

    @hybrid_property
    def flag_merge(self):
        return self.test_flags(defs.SectionFlags.SHF_MERGE)

    @hybrid_property
    def flag_strings(self):
        return self.test_flags(defs.SectionFlags.SHF_STRINGS)

    @hybrid_property
    def flag_info_link(self):
        return self.test_flags(defs.SectionFlags.SHF_INFO_LINK)

    @hybrid_property
    def flag_link_order(self):
        return self.test_flags(defs.SectionFlags.SHF_LINK_ORDER)

    @hybrid_property
    def flag_os_nonconforming(self):
        return self.test_flags(defs.SectionFlags.SHF_OS_NONCONFORMING)

    @hybrid_property
    def flag_group(self):
        return self.test_flags(defs.SectionFlags.SHF_GROUP)

    @hybrid_property
    def flag_tls(self):
        return self.test_flags(defs.SectionFlags.SHF_TLS)

    @hybrid_property
    def flag_ordered(self): # Solaris only.
        return self.test_flags(defs.SectionFlags.SHF_ORDERED)

    @hybrid_property
    def flag_exclude(self): # Solaris only.
        return self.test_flags(defs.SectionFlags.SHF_EXCLUDE)


class Elf_Symbol(Base, RidMixIn):
    """

    Attributes
    ----------
    All attributes prefixed with 'st_' are raw values from symboltables (please refer to ELF specification);
    all other exist for your convenience.

    section_name: str

    symbol_name: str

    hidden: bool

    weak: bool

    undefined: bool
    """
    st_name = StdInteger()
    st_value = StdInteger(index = True)
    st_size = StdInteger()
    st_bind = StdInteger(index = True)
    st_type = StdInteger(index = True)
    st_other = StdInteger()
    st_shndx = StdInteger()

    section_name = Column(types.VARCHAR, index = True)
    symbol_name = Column(types.VARCHAR, index = True)
    access = StdInteger(index = True)

    @hybrid_property
    def hidden(self):
        return self.st_other in (defs.SymbolVisibility.STV_HIDDEN, defs.SymbolVisibility.STV_INTERNAL)

    @hidden.expression
    def hidden(self):
        return self.st_other.in_((defs.SymbolVisibility.STV_HIDDEN, defs.SymbolVisibility.STV_INTERNAL))

    @hybrid_property
    def weak(self):
        return self.symbol_bind() == defs.SymbolBinding.STB_WEAK

    @hybrid_property
    def absolute(self):
        return self.st_shndx == defs.SectionName.SHN_ABS

    @hybrid_property
    def undefined(self):
        return self.st_shndx == defs.SectionName.SHN_UNDEF

    @hybrid_property
    def local(self):
        return self.symbol_bind() == defs.SymbolBinding.STB_LOCAL

    @hybrid_property
    def global_(self):
        return self.symbol_bind() == defs.SymbolBinding.STB_GLOBAL

    @hybrid_method
    def get_access(self):
        return self.access

    @hybrid_method
    def test_access(self, mask):
        return self.get_access() & mask == mask

    @test_access.expression
    def test_access(cls, mask):
        return cls.get_access().op("&")(mask) == mask

    @hybrid_property
    def writeable(self):
        return self.test_access(defs.SectionFlags.SHF_WRITE)

    @hybrid_property
    def executeable(self):
        return self.test_access(defs.SectionFlags.SHF_EXECINSTR)

    @hybrid_property
    def allocate(self):
        return self.test_access(defs.SectionFlags.SHF_ALLOC)

    @hybrid_property
    def symbol_bind(self):
        return defs.SymbolBinding(self.st_bind)

    @hybrid_property
    def symbol_type(self):
        return defs.SymbolType(self.st_type)


class Elf_Comment(Base, RidMixIn):
    """
    """

    text = Column(types.VARCHAR)


class Elf_Note(Base, RidMixIn):
    """
    """

    section_name = Column(types.VARCHAR, unique = True)
    type = StdInteger()
    name = Column(types.VARCHAR)
    desc = Column(types.VARCHAR)


def calculateCacheSize(value):
    return -(value // PAGE_SIZE)

REGEXER_CACHE = {}

def regexer(value, expr):
    if not REGEXER_CACHE.get(expr):
        REGEXER_CACHE[expr] = re.compile(expr, re.UNICODE)
    re_expr = REGEXER_CACHE[expr]
    return re_expr.match(value) is not None


@event.listens_for(Engine, "connect")
def set_sqlite3_pragmas(dbapi_connection, connection_record):
    dbapi_connection.create_function("REGEXP", 2, regexer)
    cursor = dbapi_connection.cursor()
    #cursor.execute("PRAGMA jornal_mode=WAL")
    cursor.execute("PRAGMA FOREIGN_KEYS=ON")
    cursor.execute("PRAGMA PAGE_SIZE={}".format(PAGE_SIZE))
    cursor.execute("PRAGMA CACHE_SIZE={}".format(calculateCacheSize(CACHE_SIZE * 1024 * 1024)))
    cursor.execute("PRAGMA SYNCHRONOUS=OFF") # FULL
    cursor.execute("PRAGMA LOCKING_MODE=EXCLUSIVE") # NORMAL
    cursor.execute("PRAGMA TEMP_STORE=MEMORY")  # FILE
    cursor.close()


class Model(object):

    def __init__(self, debug = False):
        self._engine = create_engine("sqlite:///:memory:", echo = debug,
        connect_args={'detect_types': sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES},
        native_datetime = True)

        self._session = orm.Session(self._engine, autoflush = True, autocommit = False)

        self._metadata = Base.metadata
        Base.metadata.create_all(self.engine)
        self.session.flush()
        self.session.commit()

    @property
    def engine(self):
        return self._engine

    @property
    def session(self):
        return self._session

    @property
    def metadata(self):
        return self._metadata




from objutils.elf.arm import attributes

ATMEL_SAMD21 = b"A'\x00\x00\x00aeabi\x00\x01\x1d\x00\x00\x00\x056S-M\x00\x06\x0c\x07M\t\x01\x12\x04\x14\x01\x15\x01\x17\x03\x18\x01\x1a\x01"
ARM_ATTRS1 = b'A,\x00\x00\x00aeabi\x00\x01"\x00\x00\x00\x05ARM v6K\x00\x06\t\x07M\x08\x01\t\x01\x12\x04\x14\x01\x15\x01\x17\x03\x18\x01\x1a\x01'

ARM_ATTRS2 = bytes.fromhex("4138000000616561626900012e00000005372d4100060a0741080109020a030c01110212041301140115011703180119011a021c0122012601")

"""
Attribute Section: aeabi
File Attributes
  Tag_CPU_name: "6S-M"
  Tag_CPU_arch: v6S-M
  Tag_CPU_arch_profile: Microcontroller
  Tag_THUMB_ISA_use: Thumb-1
  Tag_ABI_PCS_wchar_t: 4
  Tag_ABI_FP_denormal: Needed
  Tag_ABI_FP_exceptions: Needed
  Tag_ABI_FP_number_model: IEEE 754
  Tag_ABI_align_needed: 8-byte
  Tag_ABI_enum_size: small
"""

"""
  Tag_CPU_name: "7-A"
  Tag_CPU_arch: v7
  Tag_CPU_arch_profile: Application
  Tag_ARM_ISA_use: Yes
  Tag_THUMB_ISA_use: Thumb-2
  Tag_FP_arch: VFPv3
  Tag_Advanced_SIMD_arch: NEONv1
  Tag_ABI_PCS_GOT_use: GOT-indirect
  Tag_ABI_PCS_wchar_t: 4
  Tag_ABI_FP_rounding: Needed
  Tag_ABI_FP_denormal: Needed
  Tag_ABI_FP_exceptions: Needed
  Tag_ABI_FP_number_model: IEEE 754
  Tag_ABI_align_needed: 8-byte
  Tag_ABI_align_preserved: 8-byte, except leaf SP
  Tag_ABI_enum_size: int
  Tag_ABI_VFP_args: VFP registers
  Tag_CPU_unaligned_access: v6
  Tag_ABI_FP_16bit_format: IEEE 754

"""

def test_arm_attrs2():
    res = attributes.parse(ARM_ATTRS2)
    assert len(res.keys()) == 1
    assert "aeabi" in res
    attrs = res['aeabi']
    assert len(attrs) == 19
    a0, a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11, a12, a13, a14, a15, a16, a17, a18 = attrs

    assert a0.tag == 5
    assert a0.tag_name == "Tag_CPU_name"
    assert a0.value == "7-A"
    assert a0.description == "7-A"

    assert a1.tag == 6
    assert a1.tag_name == "Tag_CPU_arch"
    assert a1.value == 10
    assert a1.description == "ARM v7"

    assert a2.tag == 7
    assert a2.tag_name == "Tag_CPU_arch_profile"
    assert a2.value == 65
    assert a2.description == "The application profile"

    assert a3.tag == 8
    assert a3.tag_name == "Tag_ARM_ISA_use"
    assert a3.value == 1
    assert a3.description == "The user intended that this entity could use ARM instructions"

    assert a4.tag == 9
    assert a4.tag_name == "Tag_THUMB_ISA_use"
    assert a4.value == 2
    assert a4.description == "32-bit Thumb instructions were permitted"

    assert a5.tag == 10
    assert a5.tag_name == "Tag_FP_arch"
    assert a5.value == 3
    assert a5.description == "Use of the v3 FP ISA was permitted (implies use of the v2 FP ISA) "

    assert a6.tag == 12
    assert a6.tag_name == "Tag_Advanced_SIMD_arch"
    assert a6.value == 1
    assert a6.description == "Use of the Advanced SIMDv1 Architecture (Neon) was permitted"

    assert a7.tag == 17
    assert a7.tag_name == "Tag_ABI_PCS_GOT_use"
    assert a7.value == 2
    assert a7.description == "The user permitted this entity to address imported data indirectly (e.g. via a GOT)"

    assert a8.tag == 18
    assert a8.tag_name == "Tag_ABI_PCS_wchar_t"
    assert a8.value == 4
    assert a8.description == "The user intended the size of wchar_t to be 4"

    assert a9.tag == 19
    assert a9.tag_name == "Tag_ABI_FP_rounding"
    assert a9.value == 1
    assert a9.description == "The user permitted this code to choose the IEEE 754 rounding mode at run time"

    assert a10.tag == 20
    assert a10.tag_name == "Tag_ABI_FP_denormal"
    assert a10.value == 1
    assert a10.description == "The user permitted this code to choose the IEEE 754 rounding mode at run time"

    assert a11.tag == 21
    assert a11.tag_name == "Tag_ABI_FP_exceptions"
    assert a11.value == 1
    assert a11.description == "The user permitted this code to check the IEEE 754 inexact exception"

    assert a12.tag == 23
    assert a12.tag_name == "Tag_ABI_FP_number_model"
    assert a12.value == 3
    assert a12.description == "The user permitted this code to use all the IEEE 754-defined FP encodings"

    assert a13.tag == 24
    assert a13.tag_name == "Tag_ABI_align_needed"
    assert a13.value == 1
    assert a13.description == "Code was permitted to depend on the 8-byte alignment of 8-byte data items"

    assert a14.tag == 25
    assert a14.tag_name == "Tag_ABI_align8_preserved"
    assert a14.value == 1
    assert a14.description == "Code was required to preserve 8-byte alignment of 8-byte data objects"

    assert a15.tag == 26
    assert a15.tag_name == "Tag_ABI_enum_size"
    assert a15.value == 2
    assert a15.description == "The user intended Enum containers to be 32-bit "

    assert a16.tag == 28
    assert a16.tag_name == "Tag_ABI_VFP_args"
    assert a16.value == 1
    assert a16.description == "The user intended FP parameter/result passing to conform to AAPCS, VFP variant"

    assert a17.tag == 34
    assert a17.tag_name == "Tag_CPU_unaligned_access"
    assert a17.value == 1
    assert a17.description == "The user intended that this entity might make v6-style unaligned data accesses"

    assert a18.tag == 38
    assert a18.tag_name == "Tag_ABI_FP_16bit_format"
    assert a18.value == 1
    assert a18.description == "Use of IEEE 754 (draft, November 2006) format 16-bit FP numbers was permitted "




#!/usr/bin/env python
# -*- coding: utf-8 -*-

from objutils.image import Image
from objutils.utils import PYTHON_VERSION
import unittest

if PYTHON_VERSION.major == 2:
    RESULT = "Section(address = 0X00000000, length = 10000, data = '\\x00\\x01\\x02\\x03\\x04\\x05\\x06\\...\\x07\\x08\\t\\n\\x0b\\x0c\\r\\x0e\\x0f')"
else:
    RESULT = "Section(address = 0X00000000, length = 10000, data = b'\\x00\\x01\\x02\\x03\\x04\\x05\\x06...\\x07\\x08\\t\\n\\x0b\\x0c\\r\\x0e\\x0f')"


class TestRepr(unittest.TestCase):

    def testImageRepresentation(self):
        image = Image()
        image.insert_section([x % 256 for x in range(10000)])
        image.join_sections()
        self.assertEqual(repr(image), RESULT)

if __name__ == '__main__':
    unittest.main()




#!/usr/bin/env python
# -*- coding: utf-8 -*-

__version__ = "0.1.0"

__copyright__ = """
    objutils - Object file library for Python.

   (C) 2010-2019 by Christoph Schueler <cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

from collections import namedtuple, OrderedDict

from objutils.utils import SingletonBase

class CodecDoesNotExistError(Exception): pass
class CodecAlreadyExistError(Exception): pass

Codec = namedtuple("Codec", "Reader Writer description")

class Registry(SingletonBase):

    def __init__(self):
        self._codecs = OrderedDict()

    def __iter__(self):
        return iter(self._codecs.items())

    def _get_codecs(self):
        return self._codecs

    def _get_formats(self):
        return sorted(self.codecs.keys())

    def get(self, name):
        codec = self.codecs.get(name.lower())
        if not codec:
            raise CodecDoesNotExistError(name)
        return codec

    def register(self, name, readerClass, writerClass, description = ''):
        if name in self.codecs:
            raise CodecAlreadyExistError(name)
        self._codecs[name] = Codec(readerClass, writerClass, description)
        readerClass.codecName = name
        writerClass.codecName = name

    codecs = property(_get_codecs)
    formats = property(_get_formats)

registry = Registry()
